"""Perform functional tests merge of configuration levels"""
from contextlib import contextmanager
import copy
import datetime
import os
import random

import pytest
import yaml

from orion.client import get_experiment
import orion.core
import orion.core.cli
import orion.core.evc.conflicts
from orion.core.io.database.pickleddb import PickledDB
import orion.core.io.resolve_config
from orion.core.utils.singleton import SingletonNotInstantiatedError, update_singletons
import orion.core.worker
from orion.storage.base import get_storage
from orion.storage.legacy import Legacy
from orion.testing.state import OrionState


class ConfigurationTestSuite():
    """Test suite for the configuration groups"""

    database = {}

    default_storage = {
        'type': 'legacy',
        'database': {
            'type': 'pickleddb',
            'host': 'experiment.pkl'
        }
    }

    @contextmanager
    def setup_global_config(self, tmp_path):
        """Setup temporary yaml file for the global configuration"""
        with OrionState(storage=self.default_storage):
            conf_file = tmp_path / "config.yaml"
            conf_file.write_text(yaml.dump(self.config))
            conf_files = orion.core.DEF_CONFIG_FILES_PATHS
            orion.core.DEF_CONFIG_FILES_PATHS = [conf_file]
            orion.core.config = orion.core.build_config()
            try:
                yield conf_file
            finally:
                orion.core.DEF_CONFIG_FILES_PATHS = conf_files
                orion.core.config = orion.core.build_config()

    @contextmanager
    def setup_env_var_config(self, tmp_path):
        """Setup environment variables with temporary values"""
        with self.setup_global_config(tmp_path):
            tmp = {}
            for key, value in self.env_vars.items():
                tmp[key] = os.environ.pop(key, None)
                os.environ[key] = str(value)
            try:
                yield
            finally:
                for key, value in tmp.items():
                    if value:
                        os.environ[key] = str(value)
                    else:
                        del os.environ[key]

    @contextmanager
    def setup_db_config(self, tmp_path):
        """Setup database with temporary data"""
        with self.setup_env_var_config(tmp_path):
            storage = get_storage()
            storage.create_experiment(self.database)
            yield

    @contextmanager
    def setup_local_config(self, tmp_path):
        """Setup local configuration on top"""
        with self.setup_db_config(tmp_path):
            conf_file = tmp_path / "local.yaml"
            conf_file.write_text(yaml.dump(self.local))
            yield conf_file

    @contextmanager
    def setup_cmd_args_config(self, tmp_path):
        """Setup cmd args configuration... do nothing actually?"""
        with self.setup_local_config(tmp_path) as conf_file:
            yield conf_file

    def test_global_config(self, tmp_path, monkeypatch):
        """Test that global configuration is set properly based on global yaml"""
        update_singletons()
        self.sanity_check()
        with self.setup_global_config(tmp_path):
            self.check_global_config(tmp_path, monkeypatch)

    def test_env_var_config(self, tmp_path, monkeypatch):
        """Test that env vars are set properly in global config"""
        update_singletons()
        self.sanity_check()
        with self.setup_env_var_config(tmp_path):
            self.check_env_var_config(tmp_path, monkeypatch)

    @pytest.mark.usefixtures('with_user_userxyz')
    def test_db_config(self, tmp_path):
        """Test that exp config in db overrides global config"""
        update_singletons()
        self.sanity_check()
        with self.setup_db_config(tmp_path):
            self.check_db_config()

    @pytest.mark.usefixtures('with_user_userxyz')
    def test_local_config(self, tmp_path, monkeypatch):
        """Test that local config overrides db/global config"""
        update_singletons()
        self.sanity_check()
        with self.setup_local_config(tmp_path) as conf_file:
            self.check_local_config(tmp_path, conf_file, monkeypatch)

    @pytest.mark.usefixtures('with_user_userxyz')
    def test_cmd_args_config(self, tmp_path, monkeypatch):
        """Test that cmd_args config overrides local config"""
        update_singletons()
        self.sanity_check()
        with self.setup_cmd_args_config(tmp_path) as conf_file:
            self.check_cmd_args_config(tmp_path, conf_file, monkeypatch)


class TestStorage(ConfigurationTestSuite):
    """Test suite for storage configuration"""

    config = {
        'storage': {
            'type': 'legacy',
            'database': {
                'name': 'test_name',
                'type': 'pickleddb',
                'host': 'here.pkl',
                'port': 101}}}

    env_vars = {
        'ORION_STORAGE_TYPE': 'legacy',
        'ORION_DB_NAME': 'test_env_var_name',
        'ORION_DB_TYPE': 'pickleddb',
        'ORION_DB_ADDRESS': 'there.pkl',
        'ORION_DB_PORT': '103'}

    local = {
        'storage': {
            'type': 'legacy',
            'database': {
                'type': 'pickleddb',
                'host': 'local.pkl'}}}

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.storage.to_dict() != self.config['storage']

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        update_singletons()

        assert orion.core.config.storage.to_dict() == self.config['storage']

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == 'here.pkl'

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        update_singletons()

        assert orion.core.config.storage.to_dict() == {
            'type': self.env_vars['ORION_STORAGE_TYPE'],
            'database': {
                'name': self.env_vars['ORION_DB_NAME'],
                'type': self.env_vars['ORION_DB_TYPE'],
                'host': self.env_vars['ORION_DB_ADDRESS'],
                'port': int(self.env_vars['ORION_DB_PORT'])}}

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == self.env_vars['ORION_DB_ADDRESS']

    def check_db_config(self):
        """No Storage config in DB, no test"""
        pass

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        update_singletons()

        assert orion.core.config.storage.to_dict() == {
            'type': self.env_vars['ORION_STORAGE_TYPE'],
            'database': {
                'name': self.env_vars['ORION_DB_NAME'],
                'type': self.env_vars['ORION_DB_TYPE'],
                'host': self.env_vars['ORION_DB_ADDRESS'],
                'port': int(self.env_vars['ORION_DB_PORT'])}}

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test -c {conf_file} python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == 'local.pkl'

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """No Storage config in cmdline, no test"""
        pass


class TestDatabaseDeprecated(ConfigurationTestSuite):
    """Test suite for deprecated database configuration."""

    config = {
        'database': {
            'name': 'test_name',
            'type': 'pickleddb',
            'host': 'dbhere.pkl',
            'port': 101}}

    env_vars = {
        'ORION_DB_NAME': 'test_env_var_name',
        'ORION_DB_TYPE': 'pickleddb',
        'ORION_DB_ADDRESS': 'there.pkl',
        'ORION_DB_PORT': '103'}

    local = {
        'database': {
            'type': 'pickleddb',
            'host': 'dblocal.pkl'}}

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.database.to_dict() != self.config['database']

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        update_singletons()

        assert orion.core.config.database.to_dict() == self.config['database']

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == 'dbhere.pkl'

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        update_singletons()

        assert orion.core.config.database.to_dict() == {
            'name': self.env_vars['ORION_DB_NAME'],
            'type': self.env_vars['ORION_DB_TYPE'],
            'host': self.env_vars['ORION_DB_ADDRESS'],
            'port': int(self.env_vars['ORION_DB_PORT'])}

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == self.env_vars['ORION_DB_ADDRESS']

    def check_db_config(self):
        """No Storage config in DB, no test"""
        pass

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        update_singletons()

        assert orion.core.config.database.to_dict() == {
            'name': self.env_vars['ORION_DB_NAME'],
            'type': self.env_vars['ORION_DB_TYPE'],
            'host': self.env_vars['ORION_DB_ADDRESS'],
            'port': int(self.env_vars['ORION_DB_PORT'])}

        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test -c {conf_file} python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()
        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)
        assert storage._db.host == 'dblocal.pkl'

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """No Storage config in cmdline, no test"""
        pass


class TestExperimentConfig(ConfigurationTestSuite):
    """Test suite for experiment configuration"""

    config = {
        'experiment': {
            'max_trials': 10,
            'max_broken': 5,
            'working_dir': 'here',
            'pool_size': 2,
            'worker_trials': 5,
            'algorithms': {
                'aa': {
                    'b': 'c',
                    'd': {'e': 'f'}
                }
            },
            'strategy': {
                'sa': {
                    'c': 'd',
                    'e': {'f': 'g'}
                }
            }
        }
    }

    env_vars = {
        'ORION_EXP_MAX_TRIALS': 20,
        'ORION_EXP_MAX_BROKEN': 12,
        'ORION_WORKING_DIR': 'over_there'}

    database = {
        'name': 'test-name',
        'version': 1,
        'max_trials': 75,
        'max_broken': 16,
        'working_dir': 'in_db?',
        'algorithms': {
            'ab': {
                'd': 'i',
                'f': 'g'}
        },
        'producer': {
            'strategy': {
                'sb': {
                    'e': 'c',
                    'd': 'g'}
            }
        },
        'space': {
            '/x': 'uniform(0, 1)'
        },
        'metadata': {
            'VCS': {},
            'datetime': datetime.datetime.utcnow(),
            'orion_version': orion.core.__version__,
            'parser': {
                'cmd_priors': [['/x', 'uniform(0, 1)']],
                'config_file_data': {},
                'config_prefix': 'config',
                'converter': None,
                'file_config_path': None,
                'file_priors': [],
                'parser': {
                    'arguments': [
                        ['_pos_0', 'python'],
                        ['_pos_1', os.path.abspath(__file__)],
                        ['x', 'orion~uniform(0, 1)']
                    ],
                    'keys': [
                        ['_pos_0', '_pos_0'],
                        ['_pos_1', '_pos_1'],
                        ['x', '-x']
                    ],
                    'template': ['{_pos_0}', '{_pos_1}', '-x', '{x}']
                }
            },
            'priors': {'/x': 'uniform(0, 1)'},
            'user': 'userxyz',
            'user_args': ['python', os.path.abspath(__file__), '-x~uniform(0, 1)'],
            'user_script': os.path.abspath(__file__)
        },
        'refers': {'adapter': [], 'parent_id': None, 'root_id': 1}
    }

    local = {
        'experiment': {
            'name': 'test-name',
            'user': 'useruvt',
            'max_trials': 50,
            'max_broken': 15,
            'working_dir': 'here_again',
            'algorithms': {
                'ac': {
                    'd': 'e',
                    'f': 'g'}
            },
            'strategy': {
                'sd': {
                    'b': 'c',
                    'd': 'e'}
            }
        }
    }

    cmdargs = {
        'name': 'exp-name',
        'user': 'userabc',
        'version': 1,
        'exp-max-trials': 100,
        'exp-max-broken': 50,
        'working-dir': 'cmdline-working-dir',
    }

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.to_dict()['experiment'] != self.config['experiment']

    def _compare(self, base_config, experiment_config, ignore=tuple()):
        def _prune(config):
            config = copy.deepcopy(config)
            for key in ignore:
                config.pop(key, None)

            if 'producer' in config:
                config['strategy'] = config.pop('producer')['strategy']

            if 'metadata' in config and 'user' in config['metadata']:
                config['user'] = config['metadata']['user']

            return config

        experiment_config = _prune(experiment_config)
        base_config = _prune(base_config)

        for key in list(experiment_config.keys()):
            if key not in base_config:
                experiment_config.pop(key)

        assert experiment_config == base_config

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        self._compare(self.config['experiment'], orion.core.config.to_dict()['experiment'])

        script = os.path.abspath(__file__)
        command = f'hunt --init-only -n test python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()

        experiment = get_experiment('test')
        self._compare(self.config['experiment'], experiment.configuration, ignore=['worker_trials'])

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        assert orion.core.config.experiment.max_trials == self.env_vars['ORION_EXP_MAX_TRIALS']
        assert orion.core.config.experiment.max_broken == self.env_vars['ORION_EXP_MAX_BROKEN']
        assert orion.core.config.experiment.working_dir == self.env_vars['ORION_WORKING_DIR']

        script = os.path.abspath(__file__)
        command = f'hunt --init-only -n test python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        experiment = get_experiment('test')

        assert experiment.max_trials == self.env_vars['ORION_EXP_MAX_TRIALS']
        assert experiment.max_broken == self.env_vars['ORION_EXP_MAX_BROKEN']
        assert experiment.working_dir == self.env_vars['ORION_WORKING_DIR']

    def check_db_config(self):
        """Check that db config overrides global/envvar config"""
        script = os.path.abspath(__file__)
        name = 'test-name'
        command = f'hunt --worker-max-trials 0 -n {name}'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()

        experiment = get_experiment(name)
        self._compare(self.database, experiment.configuration, ignore=['worker_trials'])

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        script = os.path.abspath(__file__)
        command = f'hunt --worker-trials 0 -c {conf_file}'
        orion.core.cli.main(command.split(' '))

        storage = get_storage()

        experiment = get_experiment('test-name')
        self._compare(self.local['experiment'], experiment.configuration)

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """Check that cmdargs configuration overrides global/envvars/local configuration"""
        script = os.path.abspath(__file__)
        command = f'hunt --worker-trials 0 -c {conf_file} --branch-from test-name'
        command += ' ' + ' '.join('--{} {}'.format(name, value)
                                  for name, value in self.cmdargs.items())
        orion.core.cli.main(command.split(' '))

        storage = get_storage()

        experiment = get_experiment('exp-name')
        assert experiment.name == 'exp-name'
        assert experiment.node.parent.name == 'test-name'
        assert experiment.version == 1
        assert experiment.metadata['user'] == self.cmdargs['user']
        assert experiment.max_trials == self.cmdargs['exp-max-trials']
        assert experiment.max_broken == self.cmdargs['exp-max-broken']
        assert experiment.working_dir == self.cmdargs['working-dir']


class TestWorkerConfig(ConfigurationTestSuite):
    """Test suite for worker configuration"""

    config = {
        'worker': {
            'heartbeat': 30,
            'max_trials': 10,
            'max_broken': 5,
            'max_idle_time': 15,
            'interrupt_signal_code': 131,
            'user_script_config': 'cfg',
        }
    }

    env_vars = {
        'ORION_HEARTBEAT': 40,
        'ORION_WORKER_MAX_TRIALS': 20,
        'ORION_WORKER_MAX_BROKEN': 6,
        'ORION_MAX_IDLE_TIME': 16,
        'ORION_INTERRUPT_CODE': 132,
        'ORION_USER_SCRIPT_CONFIG': 'envcfg'}

    local = {
        'worker': {
            'heartbeat': 50,
            'max_trials': 30,
            'max_broken': 7,
            'max_idle_time': 17,
            'interrupt_signal_code': 133,
            'user_script_config': 'lclcfg',
        }
    }

    cmdargs = {
        'heartbeat': 70,
        'worker-max-trials': 40,
        'worker-max-broken': 8,
        'max-idle-time': 18,
        'interrupt-signal-code': 134,
        'user-script-config': 'cmdcfg',
    }

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.to_dict()['worker'] != self.config['worker']

    def _mock_consumer(self, monkeypatch):
        self.consumer = None
        old_init = orion.core.worker.Consumer.__init__

        def init(c_self, *args, **kwargs):
            old_init(c_self, *args, **kwargs)
            self.consumer = c_self

        monkeypatch.setattr(orion.core.worker.Consumer, '__init__', init)

    def _mock_producer(self, monkeypatch):
        self.producer = None
        old_init = orion.core.worker.Producer.__init__

        def init(p_self, *args, **kwargs):
            old_init(p_self, *args, **kwargs)
            self.producer = p_self

        monkeypatch.setattr(orion.core.worker.Producer, '__init__', init)

    def _mock_workon(self, monkeypatch):
        workon = orion.core.worker.workon

        self.workon_kwargs = None

        def mocked_workon(experiment, **kwargs):
            self.workon_kwargs = kwargs
            return workon(experiment, **kwargs)

        monkeypatch.setattr('orion.core.cli.hunt.workon', mocked_workon)

    def _check_consumer(self, config):
        assert self.consumer.heartbeat == config['heartbeat']
        assert self.consumer.template_builder.config_prefix == config['user_script_config']
        assert self.consumer.interrupt_signal_code == config['interrupt_signal_code']

    def _check_producer(self, config):
        assert self.producer.max_idle_time == config['max_idle_time']

    def _check_workon(self, config):
        assert self.workon_kwargs['max_trials'] == config['max_trials']
        assert self.workon_kwargs['max_broken'] == config['max_broken']

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        assert orion.core.config.to_dict()['worker'] == self.config['worker']

        self._mock_consumer(monkeypatch)
        self._mock_producer(monkeypatch)
        self._mock_workon(monkeypatch)

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        self._check_consumer(self.config['worker'])
        self._check_producer(self.config['worker'])
        self._check_workon(self.config['worker'])

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        env_var_config = {
            'heartbeat': self.env_vars['ORION_HEARTBEAT'],
            'max_trials': self.env_vars['ORION_WORKER_MAX_TRIALS'],
            'max_broken': self.env_vars['ORION_WORKER_MAX_BROKEN'],
            'max_idle_time': self.env_vars['ORION_MAX_IDLE_TIME'],
            'interrupt_signal_code': self.env_vars['ORION_INTERRUPT_CODE'],
            'user_script_config': self.env_vars['ORION_USER_SCRIPT_CONFIG']
        }

        assert orion.core.config.to_dict()['worker'] == env_var_config

        self._mock_consumer(monkeypatch)
        self._mock_producer(monkeypatch)
        self._mock_workon(monkeypatch)

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        self._check_consumer(env_var_config)
        self._check_producer(env_var_config)
        self._check_workon(env_var_config)

    def check_db_config(self):
        """No Storage config in DB, no test"""
        pass

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        self._mock_consumer(monkeypatch)
        self._mock_producer(monkeypatch)
        self._mock_workon(monkeypatch)

        script = os.path.abspath(__file__)
        command = f'hunt --exp-max-trials 0 -n test -c {conf_file} python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        self._check_consumer(self.local['worker'])
        self._check_producer(self.local['worker'])
        self._check_workon(self.local['worker'])

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """Check that cmdargs configuration overrides global/envvars/local configuration"""
        config = {
            'heartbeat': self.cmdargs['heartbeat'],
            'max_trials': self.cmdargs['worker-max-trials'],
            'max_broken': self.cmdargs['worker-max-broken'],
            'max_idle_time': self.cmdargs['max-idle-time'],
            'interrupt_signal_code': self.cmdargs['interrupt-signal-code'],
            'user_script_config': self.cmdargs['user-script-config'],
        }

        self._mock_consumer(monkeypatch)
        self._mock_producer(monkeypatch)
        self._mock_workon(monkeypatch)

        script = os.path.abspath(__file__)
        command = f'hunt --worker-trials 0 -c {conf_file} -n cmd-test'
        command += ' ' + ' '.join('--{} {}'.format(name, value)
                                  for name, value in self.cmdargs.items())
        command += f' python {script} -x~uniform(0,1)'
        orion.core.cli.main(command.split(' '))

        self._check_consumer(config)
        self._check_producer(config)
        self._check_workon(config)


class TestEVCConfig(ConfigurationTestSuite):
    """Test for EVC configuration"""

    config = {
        'evc': {
            'auto_resolution': False,
            'manual_resolution': True,
            'non_monitored_arguments': ['test', 'one'],
            'ignore_code_changes': True,
            'algorithm_change': True,
            'code_change_type': 'noeffect',
            'cli_change_type': 'noeffect',
            'config_change_type': 'noeffect',
        }
    }

    env_vars = {
        'ORION_EVC_MANUAL_RESOLUTION': '',
        'ORION_EVC_NON_MONITORED_ARGUMENTS': 'test:two:others',
        'ORION_EVC_IGNORE_CODE_CHANGES': '',
        'ORION_EVC_ALGO_CHANGE': '',
        'ORION_EVC_CODE_CHANGE': 'unsure',
        'ORION_EVC_CMDLINE_CHANGE': 'unsure',
        'ORION_EVC_CONFIG_CHANGE': 'unsure'
    }

    local = {
        'evc': {
            'manual_resolution': True,
            'non_monitored_arguments': ['test', 'local'],
            'ignore_code_changes': True,
            'algorithm_change': True,
            'code_change_type': 'break',
            'cli_change_type': 'break',
            'config_change_type': 'noeffect',
        }
    }

    cmdargs = {
        'manual-resolution': False,
        'non-monitored-arguments': 'test:cmdargs',
        'ignore-code-changes': False,
        'algorithm-change': False,
        'code-change-type': 'noeffect',
        'cli-change-type': 'unsure',
        'config-change-type': 'break',
    }

    def sanity_check(self):
        """Check that defaults are different than testing configuration"""
        assert orion.core.config.to_dict()['evc'] != self.config['evc']

    def _mock_consumer(self, monkeypatch):

        self.consumer = None

        def register(cls, *args, **kwargs):
            obj = super(orion.core.worker.Consumer, cls).__new__(cls)
            self.consumer = obj
            return obj

        monkeypatch.setattr(orion.core.worker.Consumer, '__new__', register)

    def _mock_producer(self, monkeypatch):

        self.producer = None

        def register(cls, *args, **kwargs):
            obj = super(orion.core.worker.Producer, cls).__new__(cls)
            self.producer = obj
            return obj

        monkeypatch.setattr(orion.core.worker.Producer, '__new__', register)

    def _mock_workon(self, monkeypatch):
        workon = orion.core.worker.workon

        self.workon_kwargs = None

        def mocked_workon(experiment, **kwargs):
            self.workon_kwargs = kwargs
            return workon(experiment, **kwargs)

        monkeypatch.setattr('orion.core.cli.hunt.workon', mocked_workon)

    def _check_consumer(self, config):

        assert self.consumer.heartbeat == config['heartbeat']
        assert self.consumer.template_builder.config_prefix == config['user_script_config']
        assert self.consumer.interrupt_signal_code == config['interrupt_signal_code']

    def _check_producer(self, config):

        assert self.producer.max_idle_time == config['max_idle_time']

    def _check_workon(self, config):

        assert self.workon_kwargs['max_trials'] == config['max_trials']
        assert self.workon_kwargs['max_broken'] == config['max_broken']

    def check_global_config(self, tmp_path, monkeypatch):
        """Check that global configuration is set properly"""
        assert orion.core.config.to_dict()['evc'] == self.config['evc']

        script = os.path.abspath(__file__)
        name = 'global-test'
        command = f'hunt --init-only -n {name} python {script} -x~uniform(0,1)'
        assert orion.core.cli.main(command.split(' ')) == 0

        # Test that manual_resolution is True and it branches when changing cli
        assert orion.core.cli.main(f'{command} --cli-change '.split(' ')) == 1

        command = 'hunt --auto-resolution ' + command[5:]

        command = self._check_cli_change(
            name, command, version=1, change_type='noeffect')
        command = self._check_non_monitored_arguments(
            name, command, version=2,
            non_monitored_arguments=['test', 'one'])
        self._check_script_config_change(
            tmp_path, name, command, version=2, change_type='noeffect')
        self._check_code_change(
            monkeypatch, name, command, version=3,
            ignore_code_change=None,
            change_type='noeffect')

    def check_env_var_config(self, tmp_path, monkeypatch):
        """Check that env vars overrides global configuration"""
        assert not orion.core.config.evc.manual_resolution
        assert not orion.core.config.evc.ignore_code_changes
        assert not orion.core.config.evc.algorithm_change
        assert orion.core.config.evc.non_monitored_arguments == ['test', 'two', 'others']
        assert orion.core.config.evc.code_change_type == 'unsure'
        assert orion.core.config.evc.cli_change_type == 'unsure'
        assert orion.core.config.evc.config_change_type == 'unsure'

        script = os.path.abspath(__file__)
        name = 'env-var-test'
        command = f'hunt --init-only -n {name} python {script} -x~uniform(0,1)'
        assert orion.core.cli.main(command.split(' ')) == 0

        # TODO: Anything to test still???
        command = self._check_cli_change(
            name, command, version=1, change_type='unsure')
        command = self._check_non_monitored_arguments(
            name, command, version=2,
            non_monitored_arguments=['test', 'two', 'others'])
        self._check_script_config_change(
            tmp_path, name, command, version=2, change_type='unsure')

        self._check_code_change(
            monkeypatch, name, command, version=3,
            ignore_code_change=None,
            change_type='unsure')

    def check_db_config(self):
        """No Storage config in DB, no test"""
        pass

    def check_local_config(self, tmp_path, conf_file, monkeypatch):
        """Check that local configuration overrides global/envvars configuration"""
        script = os.path.abspath(__file__)
        name = 'local-test'
        command = f'hunt --init-only -n {name} -c {conf_file} python {script} -x~uniform(0,1)'
        assert orion.core.cli.main(command.split(' ')) == 0

        # Test that manual_resolution is True and it branches when changing cli
        assert orion.core.cli.main(f'{command} --cli-change '.split(' ')) == 1

        command = 'hunt --auto-resolution ' + command[5:]

        command = self._check_cli_change(
            name, command, version=1, change_type=self.local['evc']['cli_change_type'])
        command = self._check_non_monitored_arguments(
            name, command, version=2,
            non_monitored_arguments=self.local['evc']['non_monitored_arguments'])
        self._check_script_config_change(
            tmp_path, name, command, version=2, change_type=self.local['evc']['config_change_type'])
        self._check_code_change(
            monkeypatch, name, command, version=3,
            ignore_code_change=True,
            change_type=self.local['evc']['code_change_type'])

    def check_cmd_args_config(self, tmp_path, conf_file, monkeypatch):
        """Check that cmdargs configuration overrides global/envvars/local configuration"""
        script = os.path.abspath(__file__)
        name = 'cmd-test'
        command = (f'hunt --init-only -c {conf_file} -n {name} '
                   '--auto-resolution '
                   '--non-monitored-arguments test:cmdargs '
                   '--code-change-type noeffect '
                   '--cli-change-type unsure '
                   '--config-change-type break '
                   f'python {script} -x~uniform(0,1)')
        assert orion.core.cli.main(command.split(' ')) == 0

        command = self._check_cli_change(
            name, command, version=1, change_type=self.cmdargs['cli-change-type'])
        command = self._check_non_monitored_arguments(
            name, command, version=2,
            non_monitored_arguments=self.cmdargs['non-monitored-arguments'].split(':'))

        self._check_script_config_change(
            tmp_path, name, command, version=2, change_type=self.cmdargs['config-change-type'])

        # Mock local to ignore_code_changes=False
        fetch_config = orion.core.io.resolve_config.fetch_config

        def mock_local(cmdargs):
            config = fetch_config(cmdargs)
            config['evc']['ignore_code_changes'] = False
            return config

        monkeypatch.setattr(orion.core.io.resolve_config, 'fetch_config', mock_local)

        # Check that ignore_code_changes is rightly False
        self._check_code_change(
            monkeypatch, name, command, version=3,
            ignore_code_change=False,
            change_type=self.cmdargs['code-change-type'])

        command = 'hunt --ignore-code-changes ' + command[5:]

        # Check that ignore_code_changes is now True
        self._check_code_change(
            monkeypatch, name, command, version=4,
            ignore_code_change=True,
            change_type=self.cmdargs['code-change-type'])

    def _check_cli_change(self, name, command, version, change_type):
        command += ' --cli-change'

        # Test that manual_resolution is False and it branches when changing cli
        assert orion.core.cli.main(command.split(' ')) == 0

        experiment = get_experiment(name, version=version + 1)
        assert experiment.version == version + 1
        assert (experiment.refers['adapter'].configuration[0] ==
                {'of_type': 'commandlinechange', 'change_type': change_type})

        return command

    def _check_non_monitored_arguments(self, name, command, version, non_monitored_arguments):
        for argument in non_monitored_arguments:
            command += f' --{argument} '

        # Test that cli change with non-monitored args do not cause branching
        assert orion.core.cli.main(command.split(' ')) == 0

        experiment = get_experiment(name, version=version + 1)
        assert experiment.version == version

        return command

    def _check_code_change(self, monkeypatch, name, command, version, ignore_code_change,
                           change_type):

        # Test that code change is handled with 'no-effect'
        def fixed_dictionary(user_script):
            """Create VCS"""
            vcs = {}
            vcs['type'] = 'git'
            vcs['is_dirty'] = False
            vcs['HEAD_sha'] = "test " + str(random.random())
            vcs['active_branch'] = None
            vcs['diff_sha'] = "diff"
            return vcs
        monkeypatch.setattr(
            orion.core.io.resolve_config, "infer_versioning_metadata", fixed_dictionary)

        detect = orion.core.evc.conflicts.CodeConflict.detect

        def mock_detect(old_config, new_config, branching_config):
            if 'ignore_code_changes' in branching_config:
                assert branching_config['ignore_code_changes'] is ignore_code_change
                branching_config['ignore_code_changes'] = False
            return detect(old_config, new_config, branching_config)

        monkeypatch.setattr(orion.core.evc.conflicts.CodeConflict, 'detect', mock_detect)
        assert orion.core.cli.main(command.split(' ')) == 0
        experiment = get_experiment(name, version=version + 1)
        assert experiment.version == version + 1
        assert (experiment.refers['adapter'].configuration[0] ==
                {'of_type': 'codechange', 'change_type': change_type})

        monkeypatch.undo()

    def _check_script_config_change(self, tmp_path, name, command, version, change_type):

        # Test that config change is handled with 'break'
        with self.setup_user_script_config(tmp_path) as user_script_config:

            command += f' --config {user_script_config}'
            assert orion.core.cli.main(command.split(' ')) == 0

        experiment = get_experiment(name, version=version + 1)
        assert experiment.version == version + 1
        print(experiment.refers['adapter'].configuration)
        assert len(experiment.refers['adapter'].configuration) == 2
        assert (experiment.refers['adapter'].configuration[1] ==
                {'of_type': 'scriptconfigchange', 'change_type': change_type})

    @contextmanager
    def setup_user_script_config(self, tmp_path):
        """Setup temporary dummy user script config"""
        conf_file = tmp_path / "user_script_config.yaml"
        config = {'what': 'ever'}
        conf_file.write_text(yaml.dump(config))
        yield conf_file



# -*- coding: utf-8 -*-
"""Tests :func:`orion.analysis.lpi`"""
import copy

import numpy
import pandas as pd
import pytest
from sklearn.ensemble import AdaBoostRegressor, BaggingRegressor,\
    ExtraTreesRegressor, GradientBoostingRegressor, RandomForestRegressor

from orion.analysis.lpi import compute_variances, lpi, make_grid, to_numpy, train_regressor
from orion.core.io.space_builder import SpaceBuilder


data = pd.DataFrame(data={
    'id': ['a', 'b', 'c', 'd'],
    'x': [0, 1, 2, 3],
    'y': [1, 2, 0, 3],
    'objective': [0.1, 0.2, 0.3, 0.5]
})


space = SpaceBuilder().build({'x': 'uniform(0, 6)', 'y': 'uniform(0, 3)'})


def test_accept_empty():
    """Tests an empty dataframe is returned if you give an empty dataframe"""
    empty_frame = pd.DataFrame()
    results = lpi(empty_frame, space)

    assert results.columns.tolist() == ['LPI']
    assert results.index.tolist() == list(space.keys())
    assert results['LPI'].tolist() == [0, 0]

    empty_frame = pd.DataFrame(columns=['x', 'y', 'objective'])
    results = lpi(empty_frame, space)

    assert results.columns.tolist() == ['LPI']
    assert results.index.tolist() == list(space.keys())
    assert results['LPI'].tolist() == [0, 0]


def test_parameter_not_modified():
    """Tests the original dataframe is not modified"""
    original = copy.deepcopy(data)
    lpi(data, space)

    pd.testing.assert_frame_equal(data, original)


def test_to_numpy():
    """Test that trials are correctly converted to numpy array"""
    array = to_numpy(data, space)

    assert array.shape == (4, 3)
    numpy.testing.assert_equal(array[:, 0], data['x'])
    numpy.testing.assert_equal(array[:, 1], data['y'])
    numpy.testing.assert_equal(array[:, 2], data['objective'])


def test_train_regressor():
    """Test training different models"""
    array = to_numpy(data, space)
    model = train_regressor('AdaBoostRegressor', array)
    assert isinstance(model, AdaBoostRegressor)
    model = train_regressor('BaggingRegressor', array)
    assert isinstance(model, BaggingRegressor)
    model = train_regressor('ExtraTreesRegressor', array)
    assert isinstance(model, ExtraTreesRegressor)
    model = train_regressor('GradientBoostingRegressor', array)
    assert isinstance(model, GradientBoostingRegressor)
    model = train_regressor('RandomForestRegressor', array)
    assert isinstance(model, RandomForestRegressor)


def test_train_regressor_kwargs():
    """Test training models with kwargs"""
    array = to_numpy(data, space)
    model = train_regressor('RandomForestRegressor', array, max_depth=2, max_features='sqrt')
    assert model.max_depth == 2
    assert model.max_features == 'sqrt'


def test_train_regressor_invalid():
    """Test error message for invalid model names"""
    array = to_numpy(data, space)
    with pytest.raises(ValueError) as exc:
        train_regressor('IDontExist', array)
    assert exc.match('IDontExist is not a supported regressor')


def test_make_grid():
    """Test grid has correct format"""
    trials = to_numpy(data, space)
    model = train_regressor('RandomForestRegressor', trials)
    best_point = trials[numpy.argmin(trials[:, -1])]
    grid = make_grid(best_point, space, model, 4)

    # Are fixed to anchor value
    numpy.testing.assert_equal(grid[0][:, 1], best_point[1])
    numpy.testing.assert_equal(grid[1][:, 0], best_point[0])

    # Is a grid in search space
    numpy.testing.assert_equal(grid[0][:, 0], [0, 2, 4, 6])
    numpy.testing.assert_equal(grid[1][:, 1], [0, 1, 2, 3])


def test_make_grid_predictor(monkeypatch):
    """Test grid contains corresponding predictions from the model"""
    trials = to_numpy(data, space)
    model = train_regressor('RandomForestRegressor', trials)
    best_point = trials[numpy.argmin(trials[:, -1])]

    # Make sure model is not predicting exactly the original objective
    with numpy.testing.assert_raises(AssertionError):
        numpy.testing.assert_equal(best_point[-1], model.predict(best_point[:-1].reshape(1, -1)))

    grid = make_grid(best_point, space, model, 4)

    # Verify that grid predictions are those of the model
    numpy.testing.assert_equal(grid[0][:, -1], model.predict(grid[0][:, :-1]))
    numpy.testing.assert_equal(grid[1][:, -1], model.predict(grid[1][:, :-1]))

    # Verify model predictions differ on different points
    with numpy.testing.assert_raises(AssertionError):
        numpy.testing.assert_equal(grid[0][:, -1], grid[1][:, -1])


def test_compute_variance():
    """Test variance computation over the grid"""
    grid = numpy.arange(3 * 5 * 4).reshape(3, 5, 4)

    grid[0, :, -1] = 10
    grid[1, :, -1] = [0, 1, 2, 3, 4]
    grid[2, :, -1] = [0, 10, 20, 30, 40]

    variances = compute_variances(grid)
    assert variances.shape == (3, )
    assert variances[0] == 0
    assert variances[1] == numpy.var([0, 1, 2, 3, 4])
    assert variances[2] == numpy.var([0, 10, 20, 30, 40])


def test_lpi_results():
    """Verify LPI results in DataFrame"""
    results = lpi(data, space, random_state=1)
    assert results.columns.tolist() == ['LPI']
    assert results.index.tolist() == list(space.keys())
    # The data is made such that x correlates more strongly with objective than y
    assert results['LPI'].loc['x'] > results['LPI'].loc['y']


def test_lpi_with_categorical_data():
    """Verify LPI can be computed on categorical dimensions"""
    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c', 'd'],
        'x': [0, 1, 2, 3],
        'y': ['b', 'c', 'a', 'd'],
        'objective': [0.1, 0.2, 0.3, 0.5]
    })

    space = SpaceBuilder().build(
        {'x': 'uniform(0, 6)',
         'y': 'choices(["a", "b", "c", "d"])'})

    results = lpi(data, space, random_state=1)
    assert results.columns.tolist() == ['LPI']
    assert results.index.tolist() == ['x', 'y']
    # The data is made such that x correlates more strongly with objective than y
    assert results['LPI'].loc['x'] > results['LPI'].loc['y']


def test_lpi_with_multidim_data():
    """Verify LPI can be computed on categorical dimensions"""
    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c', 'd'],
        'x': [[0, 2, 4], [1, 1, 3], [2, 2, 2], [3, 0, 3]],
        'y': [['b', 'b'], ['c', 'b'], ['a', 'a'], ['d', 'c']],
        'objective': [0.1, 0.2, 0.3, 0.5]
    })

    space = SpaceBuilder().build(
        {'x': 'uniform(0, 6, shape=3)',
         'y': 'choices(["a", "b", "c", "d"], shape=2)'})

    results = lpi(data, space, random_state=1)
    assert results.columns.tolist() == ['LPI']
    assert results.index.tolist() == ['x[0]', 'x[1]', 'x[2]', 'y[0]', 'y[1]']
    # The data is made such some x correlates more strongly with objective than other x and most y
    assert results['LPI'].loc['x[0]'] > results['LPI'].loc['x[1]']
    assert results['LPI'].loc['x[1]'] > results['LPI'].loc['x[2]']
    assert results['LPI'].loc['x[0]'] > results['LPI'].loc['y[0]']
    assert results['LPI'].loc['x[0]'] > results['LPI'].loc['y[1]']



# -*- coding: utf-8 -*-
"""
:mod:`orion.testing` -- Common testing support module
=====================================================
.. module:: testing
   :platform: Unix
   :synopsis: Common testing support module providing defaults, functions and mocks.
"""
# pylint: disable=protected-access

from contextlib import contextmanager
import copy
import datetime

import orion.algo.space
import orion.core.io.experiment_builder as experiment_builder
from orion.core.io.space_builder import SpaceBuilder
from orion.core.utils.format_trials import tuple_to_trial
from orion.core.worker.producer import Producer
from orion.testing.state import OrionState


def default_datetime():
    """Return default datetime"""
    return datetime.datetime(1903, 4, 25, 0, 0, 0)


def generate_trials(trial_config, statuses, exp_config=None):
    """Generate Trials with different configurations"""

    def _generate(obj, *args, value):
        if obj is None:
            return None

        obj = copy.deepcopy(obj)
        data = obj

        data[args[-1]] = value
        return obj

    new_trials = [_generate(trial_config, 'status', value=s) for s in statuses]

    for i, trial in enumerate(new_trials):
        trial['submit_time'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=i)
        if trial['status'] != 'new':
            trial['start_time'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=i)

    for i, trial in enumerate(new_trials):
        if trial['status'] == 'completed':
            trial['end_time'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=i)

    if exp_config:
        space = SpaceBuilder().build(exp_config['space'])
    else:
        space = SpaceBuilder().build({'x': 'uniform(0, 200)'})

    # make each trial unique
    for i, trial in enumerate(new_trials):
        if trial['status'] == 'completed':
            trial['results'].append({
                'name': 'loss',
                'type': 'objective',
                'value': i})

        trial_stub = tuple_to_trial(space.sample(seed=i)[0], space)
        trial['params'] = trial_stub.to_dict()['params']

    return new_trials


def mock_space_iterate(monkeypatch):
    """Force space to return seeds as samples instead of actually sampling

    This is useful for tests where we want to get params we can predict (0, 1, 2, ...)
    """
    sample = orion.algo.space.Space.sample

    def iterate(self, seed, *args, **kwargs):
        """Return the points with seed value instead of sampling"""
        points = []
        for point in sample(self, seed=seed, *args, **kwargs):
            points.append([seed] * len(point))
        return points
    monkeypatch.setattr('orion.algo.space.Space.sample', iterate)


@contextmanager
def create_experiment(exp_config=None, trial_config=None, statuses=None):
    """Context manager for the creation of an ExperimentClient and storage init"""
    if exp_config is None:
        raise ValueError("Parameter 'exp_config' is missing")
    if trial_config is None:
        raise ValueError("Parameter 'trial_config' is missing")
    if statuses is None:
        statuses = ['new', 'interrupted', 'suspended', 'reserved', 'completed']

    from orion.client.experiment import ExperimentClient

    with OrionState(experiments=[exp_config],
                    trials=generate_trials(trial_config, statuses, exp_config)) as cfg:
        experiment = experiment_builder.build(name=exp_config['name'])
        if cfg.trials:
            experiment._id = cfg.trials[0]['experiment']
        client = ExperimentClient(experiment, Producer(experiment))
        yield cfg, experiment, client

    client.close()


class MockDatetime(datetime.datetime):
    """Fake Datetime"""

    @classmethod
    def utcnow(cls):
        """Return our random/fixed datetime"""
        return default_datetime()



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.resolve_config` -- Configuration parsing and resolving
=======================================================================

.. module:: resolve_config
   :platform: Unix
   :synopsis: How does orion resolve configuration settings?

How:

 - Experiment name resolves like this:
    * cmd-arg **>** cmd-provided orion_config **>** REQUIRED (no default is given)

 - Database options resolve with the following precedence (high to low):
    * cmd-provided orion_config **>** env vars **>** default files **>** defaults

.. seealso:: :const:`ENV_VARS`, :const:`ENV_VARS_DB`


 - All other managerial, `Optimization` or `Dynamic` options resolve like this:

    * cmd-args **>** cmd-provided orion_config **>** database (if experiment name
      can be found) **>** default files

Default files are given as a list at :const:`orion.core.DEF_CONFIG_FILES_PATHS` and a
precedence is respected when building the settings dictionary:

 * default orion example file **<** system-wide config **<** user-wide config

.. note:: `Optimization` entries are required, `Dynamic` entry is optional.

"""
import copy
import getpass
import hashlib
import logging
import os

import git
from numpy import inf as infinity
import yaml

import orion
import orion.core
from orion.core.io.orion_cmdline_parser import OrionCmdlineParser
from orion.core.utils.flatten import unflatten


def is_exe(path):
    """Test whether `path` describes an executable file."""
    return os.path.isfile(path) and os.access(path, os.X_OK)


log = logging.getLogger(__name__)

################################################################################
#                 Default Settings and Environmental Variables                 #
################################################################################

# Default settings for command line arguments (option, description)
DEF_CMD_MAX_TRIALS = (infinity, 'inf/until preempted')
DEF_CMD_WORKER_TRIALS = (infinity, 'inf/until preempted')
DEF_CMD_POOL_SIZE = (1, str(1))

# list containing tuples of
# (environmental variable names, configuration keys, default values)
ENV_VARS_DB = [
    ('ORION_DB_NAME', 'name'),
    ('ORION_DB_TYPE', 'type'),
    ('ORION_DB_ADDRESS', 'host'),
    ('ORION_DB_PORT', 'port')
    ]

# TODO: Default resource from environmental (localhost)

# dictionary describing lists of environmental tuples (e.g. `ENV_VARS_DB`)
# by a 'key' to be used in the experiment's configuration dict
ENV_VARS = dict(
    database=ENV_VARS_DB
    )


def _convert_dashes(config, ref):
    """Convert dash in keys to underscores based on a reference dict.

    The reference is used to avoid converting keys in dictionary that are values
    of options.
    """
    config = copy.deepcopy(config)
    for key in list(config.keys()):
        converted_key = key.replace('-', '_')
        if converted_key in ref:
            config[converted_key] = config.pop(key)

            if all(isinstance(item[converted_key], dict) for item in [config, ref]):
                config[converted_key] = _convert_dashes(config[converted_key], ref[converted_key])

    return config


# NOTE: Silencing this pylint error for now, but seriously this function is quite horrible.
#       We'll need to clean this up at some point...
# pylint:disable=too-many-branches
def fetch_config_from_cmdargs(cmdargs):
    """Turn flat cmdargs into nested dicts like orion.core.config."""
    config_file = cmdargs.pop('config', None)
    tmp_cmdargs = copy.deepcopy(cmdargs)
    tmp_cmdargs['config'] = config_file
    cmdargs['config'] = config_file
    cmdargs = tmp_cmdargs

    cmdargs_config = {}

    if cmdargs.get('max_trials') is not None:
        log.warning(
            '--max-trials is deprecated and will be removed in v0.3. '
            'Use --exp-max-trials instead')
        cmdargs_config['experiment.max_trials'] = cmdargs.pop('max_trials')

    if cmdargs.get('worker_trials') is not None:
        log.warning(
            '--worker-trials is deprecated and will be removed in v0.3. '
            'Use --worker-max-trials instead')
        cmdargs_config['worker.max_trials'] = cmdargs.pop('worker_trials')

    mappings = dict(
        experiment=dict(
            exp_max_broken='max_broken',
            exp_max_trials='max_trials'),
        worker=dict(
            worker_max_broken='max_broken',
            worker_max_trials='max_trials'))

    mappings = dict(
        experiment=dict(
            max_broken='exp_max_broken',
            max_trials='exp_max_trials'),
        worker=dict(
            max_broken='worker_max_broken',
            max_trials='worker_max_trials'))

    global_config = orion.core.config.to_dict()

    for key in ['config', 'user_args']:
        if cmdargs.get(key) not in [False, None]:
            cmdargs_config[key] = cmdargs[key]

    for key in ['name', 'user', 'version']:
        if cmdargs.get(key) not in [False, None]:
            cmdargs_config[f'experiment.{key}'] = cmdargs[key]

    for key in ['branch_from', 'branch_to']:
        if cmdargs.get(key) not in [False, None]:
            cmdargs_config[f'evc.{key}'] = cmdargs[key]

    # Apply config at the root
    for key in ['debug']:

        # Adapt to cli arguments
        cli_key = mappings.get(key, key)

        value = cmdargs.pop(cli_key, None)
        if value is not None:
            cmdargs_config[f'{key}'] = value

    # Apply to subconfigs
    for key in ['experiment', 'worker', 'evc']:
        for subkey in global_config[key].keys():

            # Adapt to cli arguments
            cli_key = mappings.get(key, {}).get(subkey, subkey)

            value = cmdargs.pop(cli_key, None)
            if value is not None:
                cmdargs_config[f'{key}.{subkey}'] = value

    return unflatten(cmdargs_config)


def fetch_config(args):
    """Return the config inside the .yaml file if present."""
    orion_file = args.get('config')
    local_config = {}
    if orion_file:
        log.debug("Found orion configuration file at: %s", os.path.abspath(orion_file.name))
        orion_file.seek(0)
        tmp_config = yaml.safe_load(orion_file)

        global_config = orion.core.config.to_dict()

        tmp_config = _convert_dashes(tmp_config, global_config)

        # Fix deprecations first because some names are shared by experiment and worker
        max_trials = tmp_config.pop('max_trials', None)
        if max_trials is not None:
            log.warning(
                '(DEPRECATED) Option `max_trials` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nexperiment:\n  max_trials: %s', max_trials)
            local_config['experiment.max_trials'] = max_trials

        worker_trials = tmp_config.get('experiment', {}).pop('worker_trials', None)
        if worker_trials is not None:
            log.warning(
                '(DEPRECATED) Option `experiment.worker_trials` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nworker:\n  max_trials: %s', worker_trials)
            local_config['worker.max_trials'] = worker_trials

        worker_trials = tmp_config.pop('worker_trials', None)
        if worker_trials is not None:
            log.warning(
                '(DEPRECATED) Option `worker_trials` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nworker:\n  max_trials: %s', worker_trials)
            local_config['worker.max_trials'] = worker_trials

        producer = tmp_config.pop('producer', None)
        if producer is not None:
            log.warning(
                '(DEPRECATED) Option `producer` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nexperiment:\n  strategy: %s', producer['strategy'])
            local_config['experiment.strategy'] = producer['strategy']

        producer = tmp_config.get('experiment', {}).pop('producer', None)
        if producer is not None:
            log.warning(
                '(DEPRECATED) Option `experiment.producer` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nexperiment:\n  strategy: %s', producer['strategy'])
            local_config['experiment.strategy'] = producer['strategy']

        local_config = unflatten(local_config)

        # For backward compatibility
        for key in ['storage', 'experiment', 'worker', 'evc']:
            subkeys = list(global_config[key].keys())

            # Arguments that are only supported locally
            if key == 'experiment':
                subkeys += ['name', 'version', 'user']
            elif key == 'evc':
                subkeys += ['branch_from', 'branch_to']

            for subkey in subkeys:
                # Backward compatibility
                backward_value = tmp_config.pop(subkey, None)
                if backward_value is not None:
                    log.warning(
                        '(DEPRECATED) Option `%s` and will be removed in v0.3. '
                        'Use instead the option' '\n%s:\n  %s:\n    %s',
                        subkey, key, subkey, yaml.dump(backward_value, indent=6))
                value = tmp_config.get(key, {}).pop(subkey, backward_value)
                if value is not None:
                    local_config.setdefault(key, {})
                    local_config[key][subkey] = value

    return local_config


def fetch_env_vars():
    """Fetch environmental variables related to orion's managerial data."""
    env_vars = {}

    for signif, evars in ENV_VARS.items():
        env_vars[signif] = {}

        for var_name, key in evars:
            value = os.getenv(var_name)

            if value is not None:
                env_vars[signif][key] = value

    return env_vars


def fetch_metadata(user=None, user_args=None, user_script_config=None):
    """Infer rest information about the process + versioning"""
    metadata = {'user': user if user else getpass.getuser()}

    metadata['orion_version'] = orion.core.__version__

    if user_args is None:
        user_args = []

    # Trailing white space are catched by argparse as an empty argument
    if len(user_args) == 1 and user_args[0] == '':
        user_args = []

    if user_script_config is None:
        user_script_config = orion.core.config.worker.user_script_config

    cmdline_parser = OrionCmdlineParser(user_script_config)
    cmdline_parser.parse(user_args)

    if cmdline_parser.user_script:
        # TODO: Remove this, it is all in cmdline_parser now
        metadata['user_script'] = cmdline_parser.user_script
        metadata['VCS'] = infer_versioning_metadata(cmdline_parser.user_script)

    if user_args:
        metadata["user_args"] = user_args
        metadata["parser"] = cmdline_parser.get_state_dict()
        metadata["priors"] = dict(cmdline_parser.priors)

    return metadata


def merge_configs(*configs, differentiators=('type', )):
    """Merge configuration dictionaries following the given hierarchy

    Suppose function is called as merge_configs(A, B, C). Then any pair (key, value) in C would
    overwrite any previous value from A or B. Same apply for B over A.

    If for some pair (key, value), the value is a dictionary, then it will either overwrite previous
    value if it was not also a directory, or it will be merged following
    `merge_configs(old_value, new_value)`.

    .. warning:

        Redefinition of subdictionaries may lead to confusing results because merges do not remove
        data.

        If for instance, we have {'a': {'b': 1, 'c': 2}} and we would like to update `'a'` such that
        it only have `{'c': 3}`, it won't work with {'a': {'c': 3}}.

        merge_configs({'a': {'b': 1, 'c': 2}}, {'a': {'c': 3}}) -> {'a': {'b': 1, 'c': 3}}

    Examples
    --------
    .. code-block:: python
        :linenos:

        a = {'a': 1, 'b': {'c': 2}}
        b = {'b': {'c': 3}}
        c = {'b': {'c': {'d': 4}}}

        m = resolve_config.merge_configs(a, b, c)

        assert m == {'a': 1, 'b': {'c': {'d': 4}}}

        a = {'a': 1, 'b': {'c': 2, 'd': 3}}
        b = {'b': {'c': 4}}
        c = {'b': {'c': {'e': 5}}}

        m = resolve_config.merge_configs(a, b, c)

        assert m == {'a': 1, 'b': {'c': {'e': 5}, 'd': 3}}

    """
    merged_config = configs[0]

    def _can_be_merged(dict_a, dict_b):

        for differentiator in differentiators:
            if (dict_a.get(differentiator, None) and
                    dict_a[differentiator] != dict_b.get(differentiator, None)):
                return False

        return True

    for config_i in configs[1:]:
        for key, value in config_i.items():
            if (isinstance(value, dict) and
                    isinstance(merged_config.get(key), dict) and
                    _can_be_merged(merged_config[key], value)):
                merged_config[key] = merge_configs(merged_config[key], value,
                                                   differentiators=differentiators)
            elif value is not None:
                merged_config[key] = value

    return merged_config


def fetch_user_repo(user_script):
    """Fetch the GIT repo and its root path given user's script."""
    dir_path = os.path.dirname(os.path.abspath(user_script))
    try:
        git_repo = git.Repo(dir_path, search_parent_directories=True)
    except git.exc.InvalidGitRepositoryError:
        git_repo = None
        logging.warning('Script %s is not in a git repository. Code modification '
                        'won\'t be detected.', os.path.abspath(user_script))
    return git_repo


def infer_versioning_metadata(user_script):
    """
    Infer information about user's script versioning if available.
    Fills the following information in VCS:

    `is_dirty` shows whether the git repo is at a clean state.
    `HEAD_sha` gives the hash of head of the repo.
    `active_branch` shows the active branch of the repo.
    `diff_sha` shows the hash of the diff in the repo.

    :returns: the `VCS` but filled with above info.

    """
    git_repo = fetch_user_repo(user_script)
    if not git_repo:
        return {}
    vcs = {}
    vcs['type'] = 'git'
    vcs['is_dirty'] = git_repo.is_dirty()
    vcs['HEAD_sha'] = git_repo.head.object.hexsha
    if git_repo.head.is_detached:
        vcs['active_branch'] = None
    else:
        vcs['active_branch'] = git_repo.active_branch.name
    # The 'diff' of the current version from the latest commit
    diff = git_repo.git.diff(git_repo.head.commit.tree).encode('utf-8')
    diff_sha = hashlib.sha256(diff).hexdigest()
    vcs['diff_sha'] = diff_sha
    return vcs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.client.experiment`."""
import atexit
import copy
import datetime
import logging

import pandas.testing
import pytest

import orion.core
from orion.core.io.database import DuplicateKeyError
from orion.core.utils.exceptions import BrokenExperiment, SampleTimeout
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage
from orion.testing import create_experiment, mock_space_iterate

config = dict(
    name='supernaekei',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'tsirif',
              'orion_version': 'XYZ',
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    version=1,
    pool_size=1,
    max_trials=10,
    max_broken=5,
    working_dir='',
    algorithms={'random': {'seed': 1}},
    producer={'strategy': 'NoParallelStrategy'},
    refers=dict(
        root_id='supernaekei',
        parent_id=None,
        adapter=[])
)


base_trial = {
    'experiment': 0,
    'status': 'new',  # new, reserved, suspended, completed, broken
    'worker': None,
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [],
    'params': []
}


def compare_trials(trials_a, trials_b):
    """Compare two trials by using their configuration"""
    def to_dict(trial):
        return trial.to_dict()

    assert list(map(to_dict, trials_a)) == list(map(to_dict, trials_b))


def compare_without_heartbeat(trial_a, trial_b):
    """Compare trials configuration ommiting heartbeat"""
    trial_a_dict = trial_a.to_dict()
    trial_b_dict = trial_b.to_dict()
    trial_a_dict.pop('heartbeat')
    trial_b_dict.pop('heartbeat')
    assert trial_a_dict == trial_b_dict


def test_plot_is_defined():
    """Tests plot() method is defined"""
    with create_experiment(config, base_trial) as (_, _, client):
        assert client.plot()


def test_experiment_fetch_trials():
    """Test compliance of client and experiment `fetch_trials()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        assert len(experiment.fetch_trials()) == 5
        compare_trials(experiment.fetch_trials(), client.fetch_trials())


def test_experiment_get_trial():
    """Test compliance of client and experiment `get_trial()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        assert experiment.get_trial(uid=0) == client.get_trial(uid=0)


def test_experiment_fetch_trials_by_status():
    """Test compliance of client and experiment `fetch_trials_by_status()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        compare_trials(experiment.fetch_trials_by_status('completed'),
                       client.fetch_trials_by_status('completed'))


def test_experiment_fetch_non_completed_trials():
    """Test compliance of client and experiment `fetch_noncompleted_trials()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        compare_trials(experiment.fetch_noncompleted_trials(), client.fetch_noncompleted_trials())


def test_experiment_to_pandas():
    """Test compliance of client and experiment `to_pandas()`"""
    with create_experiment(config, base_trial) as (cfg, experiment, client):
        pandas.testing.assert_frame_equal(experiment.to_pandas(), client.to_pandas())


class TestInsert:
    """Tests for ExperimentClient.insert"""

    def test_insert_params_wo_results(self):
        """Test insertion without results without reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.insert(dict(x=100))
            assert trial.status == 'interrupted'
            assert trial.params['x'] == 100
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))

            assert client._pacemakers == {}

    def test_insert_params_with_results(self):
        """Test insertion with results without reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            timestamp = datetime.datetime.utcnow()
            trial = client.insert(dict(x=100),
                                  [dict(name='objective', type='objective', value=101)])
            assert trial.status == 'completed'
            assert trial.params['x'] == 100
            assert trial.objective.value == 101
            assert trial.end_time >= timestamp
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))
            assert client.get_trial(uid=trial.id).objective.value == 101

            assert client._pacemakers == {}

    def test_insert_params_with_results_and_reserve(self):
        """Test insertion with results and reservation"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(
                    dict(x=100),
                    [dict(name='objective', type='objective', value=101)],
                    reserve=True)

            assert 'Cannot observe a trial and reserve it' in str(exc.value)

    def test_insert_existing_params(self, monkeypatch):
        """Test that duplicated trials cannot be saved in storage"""
        mock_space_iterate(monkeypatch)
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(DuplicateKeyError) as exc:
                client.insert(dict(x=1))

            assert ('A trial with params {\'x\': 1} already exist for experiment supernaekei-v1' ==
                    str(exc.value))

            assert client._pacemakers == {}

    def test_insert_partial_params(self):
        """Test that trial with missing dimension that has a default value can be saved"""
        config_with_default = copy.deepcopy(config)
        config_with_default['space']['y'] = 'uniform(0, 10, default_value=5)'
        trial_with_default = copy.deepcopy(base_trial)
        trial_with_default['params'].append({'name': 'y', 'type': 'real', 'value': 1})
        with create_experiment(config_with_default, trial_with_default) as (_, experiment, client):
            trial = client.insert(dict(x=100))

            assert trial.status == 'interrupted'
            assert trial.params['x'] == 100
            assert trial.params['y'] == 5
            assert trial.id in set(trial.id for trial in experiment.fetch_trials())
            compare_without_heartbeat(trial, client.get_trial(uid=trial.id))

            assert client._pacemakers == {}

    def test_insert_partial_params_missing(self):
        """Test that trial with missing dimension cannot be saved"""
        config_with_default = copy.deepcopy(config)
        config_with_default['space']['y'] = 'uniform(0, 10)'
        trial_with_default = copy.deepcopy(base_trial)
        trial_with_default['params'].append({'name': 'y', 'type': 'real', 'value': 1})
        with create_experiment(config_with_default, trial_with_default) as (_, _, client):
            with pytest.raises(ValueError) as exc:
                client.insert(dict(x=1))

            assert ('Dimension y not specified and does not have a default value.' ==
                    str(exc.value))

    def test_insert_params_and_reserve(self):
        """Test that new trial is reserved properly with `reserve=True`"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.insert(dict(x=100), reserve=True)
            assert trial.status == 'reserved'
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_insert_params_fails_not_reserved(self, monkeypatch):
        """Test that failed insertion because of duplicated trials will not reserve the original
        trial
        """
        mock_space_iterate(monkeypatch)
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(DuplicateKeyError):
                client.insert(dict(x=1), reserve=True)

            assert client._pacemakers == {}

    def test_insert_bad_params(self):
        """Test that bad params cannot be registered in storage"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(dict(x='bad bad bad'))

            assert ('Dimension x value bad bad bad is outside of prior uniform(0, 200)' ==
                    str(exc.value))
            assert client._pacemakers == {}

    def test_insert_params_bad_results(self):
        """Test that results with from format cannot be saved (trial is registered anyhow)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            with pytest.raises(ValueError) as exc:
                client.insert(dict(x=100), [dict(name='objective', type='bad bad bad', value=0)])

            assert 'Given type, bad bad bad, not one of: ' in str(exc.value)
            assert client._pacemakers == {}


class TestReserve:
    """Tests for ExperimentClient.reserve"""

    def test_reserve(self):
        """Test reservation of registered trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            assert trial.status != 'reserved'
            client.reserve(trial)
            assert trial.status == 'reserved'
            assert experiment.get_trial(trial).status == 'reserved'
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_reserve_dont_exist(self):
        """Verify that unregistered trials cannot be reserved."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment='idontexist', params=cfg.trials[0]['params'])
            with pytest.raises(ValueError) as exc:
                client.reserve(trial)

            assert 'Trial {} does not exist in database.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}

    def test_reserve_reserved_locally(self, caplog):
        """Verify that a trial cannot be reserved twice locally (warning, no exception)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            assert trial.status != 'reserved'
            client.reserve(trial)
            with caplog.at_level(logging.WARNING):
                client.reserve(trial)

            assert 'Trial {} is already reserved.'.format(trial.id) == caplog.records[-1].message

            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_reserve_reserved_remotely(self):
        """Verify that a trial cannot be reserved if already reserved by another process"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            assert trial.status == 'interrupted'
            client.reserve(trial)
            remote_pacemaker = client._pacemakers.pop(trial.id)
            assert experiment.get_trial(trial).status == 'reserved'

            trial = Trial(**cfg.trials[1])
            assert trial.status == 'interrupted'
            with pytest.raises(RuntimeError) as exc:
                client.reserve(trial)

            assert 'Could not reserve trial {}.'.format(trial.id) == str(exc.value)

            assert trial.status == 'interrupted'
            assert experiment.get_trial(trial).status == 'reserved'
            assert client._pacemakers == {}
            remote_pacemaker.stop()

    def test_reserve_race_condition(self):
        """Verify that race conditions during `reserve` is detected and raises a comprehensible
        error
        """
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[0]['_id'])
            experiment.set_trial_status(trial, 'reserved')
            trial.status = 'new'  # Let's pretend it is still available

            with pytest.raises(RuntimeError) as exc:
                client.reserve(trial)

            assert 'Could not reserve trial {}.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}


class TestRelease:
    """Tests for ExperimentClient.release"""

    def test_release(self):
        """Test releasing (to interrupted)"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            client.release(trial)
            assert trial.status == 'interrupted'
            assert experiment.get_trial(trial).status == 'interrupted'
            assert trial.id not in client._pacemakers
            assert not pacemaker.is_alive()

    def test_release_status(self):
        """Test releasing with a specific status"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            client.release(trial, 'broken')
            assert trial.status == 'broken'
            assert experiment.get_trial(trial).status == 'broken'
            assert trial.id not in client._pacemakers
            assert not pacemaker.is_alive()

    def test_release_invalid_status(self):
        """Test releasing with a specific status"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = experiment.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            with pytest.raises(ValueError) as exc:
                client.release(trial, 'mouf mouf')

            assert exc.match('Given status `mouf mouf` not one of')

    def test_release_dont_exist(self, monkeypatch):
        """Verify that unregistered trials cannot be released"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment='idontexist', params=cfg.trials[1]['params'])

            def do_nada(trial):
                """Don't do anything"""
                return None
            monkeypatch.setattr(client, '_release_reservation', do_nada)

            with pytest.raises(ValueError) as exc:
                client.release(trial)

            assert 'Trial {} does not exist in database.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}

    def test_release_race_condition(self):
        """Verify that race conditions during `release` is detected and raises a comprehensible
        error
        """
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            pacemaker = client._pacemakers[trial.id]
            # Woops! Trial got failed over from another process.
            experiment.set_trial_status(trial, 'interrupted')
            trial.status = 'reserved'  # Let's pretend we don't know.

            with pytest.raises(RuntimeError) as exc:
                client.release(trial)

            assert ('Reservation for trial {} has been lost before release.'.format(trial.id)
                    in str(exc.value))
            assert client._pacemakers == {}
            assert not pacemaker.is_alive()

    def test_release_unreserved(self):
        """Verify that unreserved trials cannot be released"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]['_id'])
            with pytest.raises(RuntimeError) as exc:
                client.release(trial)

            assert ('Trial {} had no pacemakers. Was is reserved properly?'.format(trial.id) ==
                    str(exc.value))

            assert client._pacemakers == {}


class TestClose:
    """Test close method of the client"""

    def test_close_empty(self):
        """Test client can close when no trial is reserved"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            client.close()

    def test_close_with_reserved(self):
        """Test client cannot be closed if trials are reserved."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()

            with pytest.raises(RuntimeError) as exc:
                client.close()

            assert "There is still reserved trials" in str(exc.value)

            client.release(trial)

    def test_close_unregister_atexit(self, monkeypatch):
        """Test close properly unregister the atexit function"""
        def please_dont_call_me(client):
            raise RuntimeError("Please don't call me!!!")

        monkeypatch.setattr('orion.client.experiment.set_broken_trials', please_dont_call_me)

        with create_experiment(config, base_trial) as (cfg, experiment, client):
            # The registered function in atexit is called as expected
            with pytest.raises(RuntimeError) as exc:
                atexit._run_exitfuncs()

            assert "Please don't call me!!!" == str(exc.value)

            # Unregister the function
            client.close()


class TestBroken:
    """Test handling of broken trials with atexit()"""

    def test_broken_trial(self):
        """Test that broken trials are detected"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == 'reserved'

            atexit._run_exitfuncs()

            assert client._pacemakers == {}
            assert client.get_trial(trial).status == 'broken'

    def test_atexit_with_multiple_clients(self):
        """Test that each client has a separate atexit function"""
        config1 = copy.deepcopy(config)
        config2 = copy.deepcopy(config)
        config2['name'] = 'cloned'
        with create_experiment(exp_config=config1, trial_config=base_trial) as (_, _, client1):
            with create_experiment(exp_config=config2, trial_config=base_trial) as (_, _, client2):
                trial1 = client1.suggest()
                trial2 = client2.suggest()

                assert trial1.status == 'reserved'
                assert trial2.status == 'reserved'

                atexit._run_exitfuncs()

                assert client1._pacemakers == {}
                assert client2._pacemakers == {}
                assert client1.get_trial(trial1).status == 'broken'
                assert client2.get_trial(trial2).status == 'broken'

    def test_atexit_with_multiple_clients_unregister(self, monkeypatch):
        """Test that each client has a separate atexit function that can be unregistered"""
        config1 = copy.deepcopy(config)
        config2 = copy.deepcopy(config)
        config2['name'] = 'cloned'
        with create_experiment(exp_config=config1, trial_config=base_trial) as (_, _, client1):

            def please_dont_call_me(client):
                raise RuntimeError("Please don't call me!!!")

            monkeypatch.setattr('orion.client.experiment.set_broken_trials', please_dont_call_me)

            with create_experiment(exp_config=config2, trial_config=base_trial) as (_, _, client2):
                trial1 = client1.suggest()
                trial2 = client2.suggest()

                # The registered function in atexit is called as expected
                with pytest.raises(RuntimeError) as exc:
                    atexit._run_exitfuncs()

                assert "Please don't call me!!!" == str(exc.value)

                # Unregister the function
                client2.release(trial2)
                client2.close()

                # It should not be called
                atexit._run_exitfuncs()

                assert client1._pacemakers == {}
                assert client2._pacemakers == {}
                assert client1.get_trial(trial1).status == 'broken'
                assert client2.get_trial(trial2).status == 'interrupted'

    def test_interrupted_trial(self):
        """Test that interrupted trials are not set to broken"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == 'reserved'

            try:
                raise KeyboardInterrupt
            except KeyboardInterrupt as e:
                atexit._run_exitfuncs()

            assert client._pacemakers == {}
            assert client.get_trial(trial).status == 'interrupted'


class TestSuggest:
    """Tests for ExperimentClient.suggest"""

    def test_suggest(self, monkeypatch):
        """Verify that suggest reserved availabe trials."""
        mock_space_iterate(monkeypatch)
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == 'reserved'
            assert trial.params['x'] == 1

            assert len(experiment.fetch_trials()) == 5
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_suggest_new(self):
        """Verify that suggest can create, register and reserved new trials."""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            for _ in range(3):
                trial = client.suggest()
                assert trial.status == 'reserved'
                assert len(experiment.fetch_trials()) == 5
                assert client._pacemakers[trial.id].is_alive()
                client._pacemakers[trial.id].stop()

            trial = client.suggest()
            assert trial.status == 'reserved'
            assert trial.params['x'] == 57.57
            assert len(experiment.fetch_trials()) == 6

            assert client._pacemakers[trial.id].is_alive()
            for trial_id in list(client._pacemakers.keys()):
                client._pacemakers.pop(trial_id).stop()

    def test_suggest_race_condition(self, monkeypatch):
        """Verify that race conditions to register new trials is handled"""
        mock_space_iterate(monkeypatch)
        new_value = 50.

        # algo will suggest once an already existing trial
        def amnesia(num=1):
            """Suggest a new value and then always suggest the same"""
            if amnesia.count == 0:
                value = [0]
            else:
                value = [new_value]

            amnesia.count += 1

            return [value]

        amnesia.count = 0

        with create_experiment(config, base_trial,
                               statuses=['completed']) as (cfg, experiment, client):

            monkeypatch.setattr(experiment.algorithms, 'suggest', amnesia)

            assert len(experiment.fetch_trials()) == 1

            trial = client.suggest()
            assert trial.status == 'reserved'
            assert trial.params['x'] == new_value
            assert amnesia.count == 2

            assert len(experiment.fetch_trials()) == 2
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_suggest_algo_opt_out(self, monkeypatch):
        """Verify that None is returned when algo cannot sample new trials (opting opt)"""
        def opt_out(num=1):
            """Never suggest a new trial"""
            return None

        monkeypatch.setattr(orion.core.config.worker, 'max_idle_time', 0)

        with create_experiment(config, base_trial,
                               statuses=['completed']) as (cfg, experiment, client):

            monkeypatch.setattr(experiment.algorithms, 'suggest', opt_out)

            assert len(experiment.fetch_trials()) == 1

            with pytest.raises(SampleTimeout):
                client.suggest()

    def test_suggest_is_done(self):
        """Verify that completed experiments cannot suggest new trials"""
        with create_experiment(config, base_trial,
                               statuses=['completed'] * 10) as (cfg, experiment, client):

            assert len(experiment.fetch_trials()) == 10
            assert client.is_done

            assert client.suggest() is None

    def test_suggest_is_broken(self):
        """Verify that broken experiments cannot suggest new trials"""
        with create_experiment(config, base_trial,
                               statuses=['broken'] * 10) as (cfg, experiment, client):

            assert len(experiment.fetch_trials()) == 10
            assert client.is_broken

            with pytest.raises(BrokenExperiment):
                client.suggest()

    def test_suggest_is_done_race_condition(self, monkeypatch):
        """Verify that inability to suggest because is_done becomes True during produce() is
        handled.
        """
        with create_experiment(config, base_trial,
                               statuses=['completed'] * 5) as (cfg, experiment, client):
            def is_done(self):
                """Experiment is done"""
                return True

            def set_is_done():
                """Set is_done while algo is trying to suggest"""
                monkeypatch.setattr(experiment.__class__, 'is_done', property(is_done))

            monkeypatch.setattr(client._producer, 'produce', set_is_done)

            assert len(experiment.fetch_trials()) == 5
            assert not client.is_done

            assert client.suggest() is None

            assert len(experiment.fetch_trials()) == 5
            assert client.is_done

    def test_suggest_is_broken_race_condition(self, monkeypatch):
        """Verify that experiments that gets broken during local algo.suggest gets properly
        handled
        """
        with create_experiment(config, base_trial,
                               statuses=['broken'] * 1) as (cfg, experiment, client):

            def is_broken(self):
                """Experiment is broken"""
                return True

            def set_is_broken():
                """Set is_broken while algo is trying to suggest"""
                monkeypatch.setattr(experiment.__class__, 'is_broken', property(is_broken))

            monkeypatch.setattr(client._producer, 'produce', set_is_broken)

            assert len(experiment.fetch_trials()) == 1
            assert not client.is_broken

            with pytest.raises(BrokenExperiment):
                client.suggest()

            assert len(experiment.fetch_trials()) == 1
            assert client.is_broken

    def test_suggest_hierarchical_space(self):
        """Verify that suggest returns trial with proper hierarchical parameter."""
        exp_config = copy.deepcopy(config)
        exp_config['space'] = {
            'a': {'x': 'uniform(0, 10, discrete=True)'},
            'b': {'y': 'loguniform(1e-08, 1)',
                  'z': 'choices([\'voici\', \'voila\', 2])'}}
        with create_experiment(exp_config=exp_config,
                               trial_config=base_trial,
                               statuses=[]) as (cfg, experiment, client):
            trial = client.suggest()
            assert trial.status == 'reserved'
            assert len(trial.params) == 2
            assert 'x' in trial.params['a']
            assert 'y' in trial.params['b']
            assert 'z' in trial.params['b']

            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()


class TestObserve:
    """Tests for ExperimentClient.observe"""

    def test_observe(self):
        """Verify that `observe()` will update the storage"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            assert trial.results == []
            client.reserve(trial)
            assert get_storage().get_trial(trial).objective is None
            client.observe(trial, [dict(name='objective', type='objective', value=101)])
            assert get_storage().get_trial(trial).objective.value == 101

    def test_observe_unreserved(self):
        """Verify that `observe()` will fail on non-reserved trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            with pytest.raises(RuntimeError) as exc:
                client.observe(trial, [dict(name='objective', type='objective', value=101)])

            assert ('Trial {} had no pacemakers. Was is reserved properly?'.format(trial.id) ==
                    str(exc.value))

    def test_observe_dont_exist(self):
        """Verify that `observe()` will fail on non-registered trials"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(experiment='idontexist', params=cfg.trials[0]['params'])
            with pytest.raises(ValueError) as exc:
                client.observe(trial, [dict(name='objective', type='objective', value=101)])

            assert 'Trial {} does not exist in database.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}

    def test_observe_bad_results(self):
        """Verify that bad results type is detected and ValueError is raised"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = Trial(**cfg.trials[1])
            client.reserve(trial)
            with pytest.raises(ValueError) as exc:
                client.observe(trial, [dict(name='objective', type='bad bad bad', value=101)])

            assert 'Given type, bad bad bad, not one of: ' in str(exc.value)
            assert client._pacemakers[trial.id].is_alive()
            client._pacemakers.pop(trial.id).stop()

    def test_observe_race_condition(self):
        """Verify that race condition during `observe()` is detected and raised"""
        with create_experiment(config, base_trial) as (cfg, experiment, client):
            trial = client.get_trial(uid=cfg.trials[1]['_id'])
            client.reserve(trial)
            experiment.set_trial_status(trial, 'interrupted')
            trial.status = 'reserved'  # Let's pretend it is still reserved

            with pytest.raises(RuntimeError) as exc:
                client.observe(trial, [dict(name='objective', type='objective', value=101)])

            assert 'Reservation for trial {} has been lost.'.format(trial.id) == str(exc.value)
            assert client._pacemakers == {}


class TestWorkon:
    """Tests for ExperimentClient.workon"""

    def test_workon(self):
        """Verify that workon processes properly"""
        def foo(x):
            return [dict(name='result', type='objective', value=x * 2)]

        with create_experiment(config, base_trial, statuses=[]) as (cfg, experiment, client):
            client.workon(foo, max_trials=5)
            assert len(experiment.fetch_trials()) == 5
            assert client._pacemakers == {}

    def test_workon_partial(self):
        """Verify that partial is properly passed to the function"""
        def foo(x, y):
            return [dict(name='result', type='objective', value=x * 2 + y)]

        with create_experiment(config, base_trial, statuses=[]) as (cfg, experiment, client):
            client.workon(foo, max_trials=10, y=2)
            assert len(experiment.fetch_trials()) == 10
            assert client._pacemakers == {}

    def test_workon_partial_with_override(self):
        """Verify that partial is overriden by trial.params"""
        def foo(x, y):
            return [dict(name='result', type='objective', value=x * 2 + y)]

        ext_config = copy.deepcopy(config)
        ext_config['space']['y'] = 'uniform(0, 10)'

        with create_experiment(exp_config=ext_config,
                               trial_config=base_trial,
                               statuses=[]) as (cfg, experiment, client):
            default_y = 2
            assert len(experiment.fetch_trials()) == 0
            client.workon(foo, max_trials=1, y=default_y)
            assert len(experiment.fetch_trials()) == 1
            assert experiment.fetch_trials()[0].params['y'] != 2

    def test_workon_hierarchical_partial_with_override(self):
        """Verify that hierarchical partial is overriden by trial.params"""
        default_y = 2
        default_z = 'voila'

        def foo(a, b):
            assert b['y'] != default_y
            assert b['z'] == default_z
            return [dict(name='result', type='objective', value=a['x'] * 2 + b['y'])]

        ext_config = copy.deepcopy(config)
        ext_config['space'] = {
            'a': {'x': 'uniform(0, 10, discrete=True)'},
            'b': {'y': 'loguniform(1e-08, 1)'}}

        with create_experiment(exp_config=ext_config,
                               trial_config=base_trial,
                               statuses=[]) as (cfg, experiment, client):
            assert len(experiment.fetch_trials()) == 0
            client.workon(foo, max_trials=5, b={'y': default_y, 'z': default_z})
            assert len(experiment.fetch_trials()) == 5
            params = experiment.fetch_trials()[0].params
            assert len(params)
            assert 'x' in params['a']
            assert 'y' in params['b']



"""
:mod:`orion.core.plotting.backend_plotly` -- Plotly backend for plotting methods
================================================================================

.. module:: backend_plotly
   :platform: Unix
   :synopsis: Plotly backend for plotting methods
"""
import functools

import numpy
import pandas as pd
import plotly.graph_objects as go

from orion.algo.space import Categorical, Fidelity
import orion.analysis.regret


def lpi(experiment, model='RandomForestRegressor', model_kwargs=None, n=20, **kwargs):
    """Plotly implementation of `orion.plotting.lpi`"""
    if not experiment:
        raise ValueError("Parameter 'experiment' is None")

    if model_kwargs is None:
        model_kwargs = {}

    df = experiment.to_pandas()
    df = orion.analysis.lpi(df, experiment.space, model=model, n=n, **model_kwargs)

    fig = go.Figure(data=[go.Bar(x=df.index.tolist(), y=df['LPI'].tolist())])

    y_axis_label = "Local Parameter Importance (LPI)"
    fig.update_layout(title=f"LPI for experiment '{experiment.name}'",
                      xaxis_title="Hyperparameters",
                      yaxis_title=y_axis_label)

    return fig


def parallel_coordinates(experiment, order=None, colorscale='YlOrRd', **kwargs):
    """Plotly implementation of `orion.plotting.parallel_coordinates`"""
    def build_frame():
        """Builds the dataframe for the plot"""
        names = list(experiment.space.keys())

        df = experiment.to_pandas()
        df = df.loc[df['status'] == 'completed']

        df[names] = df[names].transform(functools.partial(_curate_params, space=experiment.space))

        df = _flatten_dims(df, experiment.space)

        return df

    def infer_order(order):
        """Create order if not passed, otherwised verify it"""
        if order is None:
            order = list(experiment.space.keys())
            fidelity_dims = [dim for dim in experiment.space.values() if isinstance(dim, Fidelity)]
            if fidelity_dims:
                del order[order.index(fidelity_dims[0].name)]
                order.insert(0, fidelity_dims[0].name)
        else:
            names = set(experiment.space.keys())
            any_invalid = set(order) - names
            if any_invalid:
                raise ValueError(f'Some names are invalid: {any_invalid} not in {names}')

        return order

    def get_dimension(data, name, dim):
        dim_data = dict(label=name, values=data[name])
        if isinstance(dim, Categorical):
            dim_data['tickvals'] = list(range(len(dim.categories)))
            dim_data['ticktext'] = dim.categories
        elif isinstance(dim, Fidelity):
            dim_data['range'] = (dim.low, dim.high)
        else:
            dim_data['range'] = dim.interval()
        return dim_data

    if not experiment:
        raise ValueError("Parameter 'experiment' is None")

    trial = experiment.fetch_trials_by_status('completed')[0]
    df = build_frame()

    dimensions = []
    for name in infer_order(order):
        dim = experiment.space[name]
        shape = dim.shape
        if shape:
            assert len(shape) == 1
            for i in range(shape[0]):
                name_i = f'{name}[{i}]'
                dimensions.append(get_dimension(df, name_i, dim))
        else:
            dimensions.append(get_dimension(df, name, dim))

    objective_name = trial.objective.name

    objectives = df['objective']
    omin = min(df['objective'])
    omax = max(df['objective'])

    dimensions.append(
        dict(label=objective_name, range=(omin, omax), values=objectives))

    fig = go.Figure(
        data=go.Parcoords(
            line=dict(
                color=objectives,
                colorscale=colorscale,
                showscale=True,
                cmin=omin,
                cmax=omax,
                colorbar=dict(title=objective_name)),
            dimensions=dimensions))

    fig.update_layout(title=f"Parallel Coordinates Plot for experiment '{experiment.name}'")

    return fig


def regret(experiment, order_by, verbose_hover, **kwargs):
    """Plotly implementation of `orion.plotting.regret`"""
    def build_frame():
        """Builds the dataframe for the plot"""
        df = experiment.to_pandas()

        names = list(experiment.space.keys())
        df['params'] = df[names].apply(
            _format_hyperparameters,
            args=(names, ), axis=1)

        df = df.loc[df['status'] == 'completed']
        df = df.sort_values(order_by)
        df = orion.analysis.regret(df)
        return df

    ORDER_KEYS = ['suggested', 'reserved', 'completed']

    if not experiment:
        raise ValueError("Parameter 'experiment' is None")

    if order_by not in ORDER_KEYS:
        raise ValueError(f"Parameter 'order_by' is not one of {ORDER_KEYS}")

    trial = experiment.fetch_trials_by_status('completed')[0]
    df = build_frame()

    fig = go.Figure()

    fig.add_scatter(y=df['objective'],
                    mode='markers',
                    name='trials',
                    customdata=list(zip(df['id'], df[order_by], df['params'])),
                    hovertemplate=_template_trials(verbose_hover))
    fig.add_scatter(y=df['best'],
                    mode='lines',
                    name='best-to-date',
                    customdata=list(zip(df['best_id'], df['best'])),
                    hovertemplate=_template_best())

    y_axis_label = f"{trial.objective.type.capitalize()} '{trial.objective.name}'"
    fig.update_layout(title=f"Regret for experiment '{experiment.name}'",
                      xaxis_title=f"Trials ordered by {order_by} time",
                      yaxis_title=y_axis_label)

    return fig


def _format_value(value):
    """
    Hyperparameter can have many types, sometimes they can even be lists.
    If one of the value is a float, it has to be compact.
    """
    if isinstance(value, str):
        return value

    try:
        return f"[{','.join(_format_value(x) for x in value)}]"

    except TypeError:
        if isinstance(value, float):
            return f"{value:.5G}"
        else:
            return value


def _format_hyperparameters(hyperparameters, names):
    result = ''

    for name, value in zip(names, hyperparameters):
        x = f'<br>  {name[1:]}: {_format_value(value)}'
        result += x

    return result


def _template_trials(verbose_hover):
    template = '<b>ID: %{customdata[0]}</b><br>' \
        'value: %{y}<br>' \
        'time: %{customdata[1]|%Y-%m-%d %H:%M:%S}<br>'

    if verbose_hover:
        template += 'parameters: %{customdata[2]}'

    template += '<extra></extra>'

    return template


def _template_best():
    return '<b>Best ID: %{customdata[0]}</b><br>' \
        'value: %{customdata[1]}' \
        '<extra></extra>'


def _curate_params(data, space):
    dim = space[data.name]
    if isinstance(dim, Categorical):
        data = numpy.array(data.tolist())  # To unpack lists if dim shape > 1
        shape = data.shape
        assert len(shape) <= 2
        idx = numpy.argmax(data.reshape(-1, 1) == numpy.array(dim.categories), axis=1)
        idx = idx.reshape(shape)
        if len(shape) > 1:
            return [list(idx[i]) for i in range(shape[0])]
        return idx
    return data


def _flatten_dims(data, space):
    for key, dim in space.items():
        if dim.shape:
            assert len(dim.shape) == 1

            # expand df.tags into its own dataframe
            values = data[key].apply(pd.Series)

            # rename each hp
            values = values.rename(columns=lambda x: f'{key}[{x}]')

            data = pd.concat([data[:], values[:]], axis=1)

    return data



# -*- coding: utf-8 -*-
"""
:mod:`orion.analysis.regret` -- Provide tools to calculate regret
=================================================================

.. module:: orion.analysis.regret
   :platform: Unix
   :synopsis: Provide tools to calculate regret
"""
import numpy
import pandas as pd


def regret(trials, names=('best', 'best_id')):
    """
    Calculates the regret for a collection of :class:`Trial`. The regret is calculated sequentially
    from the order of the collection.

    Parameters
    ----------
    trials: DataFrame or dict
        A dataframe of trials containing, at least, the columns 'objective' and 'id'. Or a dict
        equivalent.
    names:
        A tuple containing the names of the columns. Default is ('best', 'best-id').

    Returns
    -------
    A copy of the original dataframe with two new columns containing respectively the best value
    so far and its trial id.
    """
    if len(names) != 2:
        raise ValueError(f"`names` requires a tuple with 2 elements. {len(names)} provided.")

    df = pd.DataFrame(trials, copy=True)
    if df.empty:
        return df

    regrets_idx = get_regrets_idx(df['objective'])
    df[names[0]] = df['objective'].to_numpy()[list(regrets_idx)]
    df[names[1]] = df['id'].to_numpy()[regrets_idx]

    return df


def get_regrets_idx(x):
    """Return the indices corresponding to the cumulative minimum"""
    minima = numpy.minimum.accumulate(x)
    diff = numpy.diff(minima)
    jumps = numpy.arange(len(x))
    jumps[1:] *= (diff != 0)
    jumps = numpy.maximum.accumulate(jumps)
    return jumps



"""
:mod:`orion.plotting.base` -- Provides public plotting API
===========================================================

.. module:: base
   :platform: Unix
   :synopsis: Provides public plotting API

"""
import orion.plotting.backend_plotly as backend


def lpi(experiment, model='RandomForestRegressor', model_kwargs=None, n=20, **kwargs):
    """
    Make a bar plot to visualize the local parameter importance metric.

    For more information on the metric, see original paper at
    https://ml.informatik.uni-freiburg.de/papers/18-LION12-CAVE.pdf.

    Biedenkapp, André, et al. "Cave: Configuration assessment, visualization and evaluation."
    International Conference on Learning and Intelligent Optimization. Springer, Cham, 2018.

    Parameters
    ----------
    experiment: ExperimentClient, Experiment or ExperimentView
        The orion object containing the experiment data

    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)

        Arguments for the regressor model.
    model_kwargs: dict
        Arguments for the regressor model.
    n: int
        Number of points to compute the variances. Default is 20.
    kwargs: dict
        All other plotting keyword arguments to be passed to
        :meth:`plotly.express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided or if regressor name is invalid.

    """
    return backend.lpi(experiment, model=model, model_kwargs=model_kwargs, n=n, **kwargs)


def parallel_coordinates(experiment, order=None, **kwargs):
    """
    Make a Parallel Coordinates Plot to visualize the effect of the hyperparameters
    on the objective.

    Parameters
    ----------
    experiment: ExperimentClient, Experiment or ExperimentView
        The orion object containing the experiment data

    order: list of str or None
        Indicates the order of columns in the parallel coordinate plot. By default
        the columns are sorted alphabetically with the exception of the first column
        which is reserved for a fidelity dimension is there is one in the search space.

    kwargs: dict
        All other plotting keyword arguments to be passed to
        :meth:`plotly.express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided.

    """
    return backend.parallel_coordinates(experiment, order=order, **kwargs)


def regret(experiment, order_by='suggested', verbose_hover=True, **kwargs):
    """
    Make a plot to visualize the performance of the hyper-optimization process.

    The x-axis contain the trials and the y-axis their respective best performance.

    Parameters
    ----------
    experiment: ExperimentClient, Experiment or ExperimentView
        The orion object containing the experiment data

    order_by: str
        Indicates how the trials should be ordered. Acceptable options are below.
        See attributes of ``Trial`` for more details.

        * 'suggested': Sort by trial suggested time (default).
        * 'reserved': Sort by trial reserved time.
        * 'completed': Sort by trial completed time.

    verbose_hover: bool
        Indicates whether to display the hyperparameter in hover tooltips. True by default.

    kwargs: dict
        All other plotting keyword arguments to be passed to
        :meth:`plotly.express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided.

    """
    return backend.regret(experiment, order_by, verbose_hover, **kwargs)


PLOT_METHODS = {
    'lpi': lpi,
    'parallel_coordinates': parallel_coordinates,
    'regret': regret}


class PlotAccessor:
    """
    Make plots of ExperimentClient.

    Parameters
    ----------
    experiment : ExperimentClient
        The object for which the method is called.

    Raises
    ------
    ValueError
        If no experiment is provided.
    """

    def __init__(self, experiment):
        if not experiment:
            raise ValueError("Parameter 'experiment' is None")
        self._experiment = experiment

    def __call__(self, **kwargs):
        """
        Make different kinds of plots of ExperimentClient.

        Parameters
        ----------
        kind : str
            The kind of plot to produce:

            - 'regret' : Regret plot (default)
        """
        kind = kwargs.pop('kind', 'regret')

        if kind not in PLOT_METHODS.keys():
            raise ValueError(f"Plot of kind '{kind}' is not one of {list(PLOT_METHODS.keys())}")

        return PLOT_METHODS[kind](self._experiment, **kwargs)

    def lpi(self, **kwargs):
        """Make a bar plot of the local parameter importance metrics."""
        __doc__ = lpi.__doc__
        return self(kind="lpi", **kwargs)

    def parallel_coordinates(self, **kwargs):
        """Make a parallel coordinates plot to visualize the performance of the
        hyper-optimization process.
        """
        __doc__ = parallel_coordinates.__doc__
        return self(kind="parallel_coordinates", **kwargs)

    def regret(self, **kwargs):
        """Make a plot to visualize the performance of the hyper-optimization process."""
        __doc__ = regret.__doc__
        return self(kind="regret", **kwargs)



"""Common fixtures and utils for configuration tests."""
from orion.algo.base import (BaseAlgorithm, OptimizationAlgorithm)
from orion.core.worker.strategy import (BaseParallelStrategy, Strategy)


def __init__(self, *args, **params):
    """Initialize the parent class"""
    self.params = params
    super(type(self), self).__init__(*args, **params)


def stub(self, *args, **kwargs):
    """Just a stub..."""
    pass


def configuration(self):
    """Configuration based on self.params"""
    return {self.__class__.__name__.lower(): self.params}


for char in 'ABCDE':
    algo_class = type(
        f'A{char}',
        (BaseAlgorithm, ),
        {'suggest': stub, 'observe': stub})

    # Hack it into being discoverable
    OptimizationAlgorithm.types.append(algo_class)
    OptimizationAlgorithm.typenames.append(algo_class.__name__.lower())


for char in 'ABCDE':
    strategy_class = type(
        f'S{char}',
        (BaseParallelStrategy, ),
        {'observe': stub, '__init__': __init__})

    strategy_class.configuration = property(configuration)

    # Hack it into being discoverable
    Strategy.types.append(strategy_class)
    Strategy.typenames.append(strategy_class.__name__.lower())



"""Collection of tests for :mod:`orion.plotting.backend_plotly`."""
import copy

import pandas
import plotly
import pytest

from orion.core.worker.experiment import Experiment, ExperimentView
from orion.plotting.base import lpi, parallel_coordinates, regret
from orion.testing import create_experiment

config = dict(
    name='experiment-name',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'test-user',
              'orion_version': 'XYZ',
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir='',
    algorithms={'random': {'seed': 1}},
    producer={'strategy': 'NoParallelStrategy'},
)

trial_config = {
    'experiment': 0,
    'status': 'completed',
    'worker': None,
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [],
    'params': []
}


def assert_regret_plot(plot):
    """Checks the layout of a regret plot"""
    assert plot.layout.title.text == "Regret for experiment 'experiment-name'"
    assert plot.layout.xaxis.title.text == "Trials ordered by suggested time"
    assert plot.layout.yaxis.title.text == "Objective 'loss'"

    trace1 = plot.data[0]
    assert trace1.type == "scatter"
    assert trace1.name == "trials"
    assert trace1.mode == "markers"
    assert len(trace1.y) == 1
    assert not trace1.x

    trace2 = plot.data[1]
    assert trace2.type == "scatter"
    assert trace2.name == "best-to-date"
    assert trace2.mode == "lines"
    assert len(trace2.y) == 1
    assert not trace2.x


def mock_space(x='uniform(0, 6)', y='uniform(0, 3)', **kwargs):
    """Build a mocked space"""
    mocked_config = copy.deepcopy(config)
    mocked_config['space'] = {'x': x, 'y': y}
    mocked_config['space'].update(kwargs)
    return mocked_config


def mock_experiment(monkeypatch, x=None, y=None):
    """Mock experiment to_pandas to return given data (or default one)"""
    if x is None:
        x = [0, 1, 2, 3]
    if y is None:
        y = [1, 2, 0, 3]

    def to_pandas(self, with_evc_tree=False):
        data = pandas.DataFrame(data={
            'id': ['a', 'b', 'c', 'd'],
            'x': x,
            'y': y,
            'objective': [0.1, 0.2, 0.3, 0.5]
        })

        return data

    monkeypatch.setattr(Experiment, 'to_pandas', to_pandas)


def assert_lpi_plot(plot, dims):
    """Checks the layout of a LPI plot"""
    assert plot.layout.title.text == "LPI for experiment 'experiment-name'"
    assert plot.layout.xaxis.title.text == "Hyperparameters"
    assert plot.layout.yaxis.title.text == "Local Parameter Importance (LPI)"

    trace = plot.data[0]
    assert trace['x'] == tuple(dims)
    assert trace['y'][0] > trace['y'][1]


class TestLPI:
    """Tests the ``lpi()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            lpi(None)

    def test_returns_plotly_object(self):
        """Tests that the plotly backend returns a plotly object"""
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            plot = lpi(experiment, random_state=1)

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self, monkeypatch):
        """Tests the layout of the plot"""
        config = mock_space()
        mock_experiment(monkeypatch)
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            plot = lpi(experiment, random_state=1)

        assert_lpi_plot(plot, dims=['x', 'y'])

    def test_experiment_worker_as_parameter(self, monkeypatch):
        """Tests that ``Experiment`` is a valid parameter"""
        config = mock_space()
        mock_experiment(monkeypatch)
        with create_experiment(config, trial_config, ['completed']) as (_, experiment, _):
            plot = lpi(experiment, random_state=1)

        assert_lpi_plot(plot, dims=['x', 'y'])

    def test_experiment_view_as_parameter(self, monkeypatch):
        """Tests that ``ExperimentView`` is a valid parameter"""
        config = mock_space()
        mock_experiment(monkeypatch)
        with create_experiment(config, trial_config, ['completed']) as (_, experiment, _):
            plot = lpi(ExperimentView(experiment), random_state=1)

        assert_lpi_plot(plot, dims=['x', 'y'])

    def test_multidim(self, monkeypatch):
        """Tests that dimensions with shape > 1 are flattened properly"""
        config = mock_space(y='uniform(0, 3, shape=2)')
        mock_experiment(monkeypatch, y=[[3, 3], [2, 3], [1, 2], [0, 3]])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=['x', 'y[0]', 'y[1]'])

    def test_fidelity(self, monkeypatch):
        """Tests that fidelity is supported"""
        config = mock_space(y='fidelity(1, 200, base=3)')
        mock_experiment(monkeypatch, y=[1, 3 ** 2, 1, 3 ** 4])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=['x', 'y'])

    def test_categorical(self, monkeypatch):
        """Tests that categorical is supported"""
        config = mock_space(y='choices(["a", "b", "c"])')
        mock_experiment(monkeypatch, y=['c', 'c', 'a', 'b'])
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=['x', 'y'])

    def test_categorical_multidim(self, monkeypatch):
        """Tests that multidim categorical is supported"""
        config = mock_space(y='choices(["a", "b", "c"], shape=3)')
        mock_experiment(
            monkeypatch,
            y=[['c', 'b', 'a'], ['c', 'a', 'c'], ['a', 'b', 'a'], ['c', 'b', 'b']])

        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = lpi(experiment, model_kwargs=dict(random_state=1))

        assert_lpi_plot(plot, dims=['x', 'y[0]', 'y[1]', 'y[2]'])


class TestRegret:
    """Tests the ``regret()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            regret(None)

    def test_returns_plotly_object(self):
        """Tests that the plotly backend returns a plotly object"""
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            plot = regret(experiment)

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self):
        """Tests the layout of the plot"""
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_experiment_worker_as_parameter(self):
        """Tests that ``Experiment`` is a valid parameter"""
        with create_experiment(config, trial_config, ['completed']) as (_, experiment, _):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_experiment_view_as_parameter(self):
        """Tests that ``ExperimentView`` is a valid parameter"""
        with create_experiment(config, trial_config, ['completed']) as (_, experiment, _):
            plot = regret(ExperimentView(experiment))

        assert_regret_plot(plot)

    def test_ignore_uncompleted_statuses(self):
        """Tests that uncompleted statuses are ignored"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_unsupported_order_key(self):
        """Tests that unsupported order keys are rejected"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            with pytest.raises(ValueError):
                regret(experiment, order_by='unsupported')


def assert_parallel_coordinates_plot(plot, order):
    """Checks the layout of a parallel coordinates plot"""
    assert plot.layout.title.text == "Parallel Coordinates Plot for experiment 'experiment-name'"

    trace = plot.data[0]
    for i in range(len(order)):
        assert trace.dimensions[i].label == order[i]


class TestParallelCoordinates:
    """Tests the ``parallel_coordinates()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            parallel_coordinates(None)

    def test_returns_plotly_object(self):
        """Tests that the plotly backend returns a plotly object"""
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self):
        """Tests the layout of the plot"""
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=['x', 'loss'])

    def test_experiment_worker_as_parameter(self):
        """Tests that ``Experiment`` is a valid parameter"""
        with create_experiment(config, trial_config, ['completed']) as (_, experiment, _):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=['x', 'loss'])

    def test_experiment_view_as_parameter(self):
        """Tests that ``ExperimentView`` is a valid parameter"""
        with create_experiment(config, trial_config, ['completed']) as (_, experiment, _):
            plot = parallel_coordinates(ExperimentView(experiment))

        assert_parallel_coordinates_plot(plot, order=['x', 'loss'])

    def test_ignore_uncompleted_statuses(self):
        """Tests that uncompleted statuses are ignored"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=['x', 'loss'])

    def test_unsupported_order_key(self):
        """Tests that unsupported order keys are rejected"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            with pytest.raises(ValueError):
                parallel_coordinates(experiment, order=['unsupported'])

    def test_order_columns(self):
        """Tests that columns are sorted according to ``order``"""
        multidim_config = copy.deepcopy(config)
        for k in 'yzutv':
            multidim_config['space'][k] = 'uniform(0, 200)'
        with create_experiment(multidim_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment, order='vzyx')

        assert_parallel_coordinates_plot(plot, order=['v', 'z', 'y', 'x', 'loss'])

    def test_multidim(self):
        """Tests that dimensions with shape > 1 are flattened properly"""
        multidim_config = copy.deepcopy(config)
        multidim_config['space']['y'] = 'uniform(0, 200, shape=4)'
        with create_experiment(multidim_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=['x', 'y[0]', 'y[1]', 'y[2]', 'y[3]', 'loss'])

    def test_fidelity(self):
        """Tests that fidelity is set to first column by default"""
        fidelity_config = copy.deepcopy(config)
        fidelity_config['space']['z'] = 'fidelity(1, 200, base=3)'
        with create_experiment(fidelity_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=['z', 'x', 'loss'])

    def test_categorical(self):
        """Tests that categorical is supported"""
        categorical_config = copy.deepcopy(config)
        categorical_config['space']['z'] = 'choices(["a", "b", "c"])'
        with create_experiment(categorical_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=['x', 'z', 'loss'])

    def test_categorical_multidim(self):
        """Tests that multidim categorical is supported"""
        categorical_config = copy.deepcopy(config)
        categorical_config['space']['z'] = 'choices(["a", "b", "c"], shape=3)'
        with create_experiment(categorical_config, trial_config) as (_, _, experiment):
            plot = parallel_coordinates(experiment)

        assert_parallel_coordinates_plot(plot, order=['x', 'z[0]', 'z[1]', 'z[2]', 'loss'])



# -*- coding: utf-8 -*-
"""
:mod:`orion.analysis.lpi` -- Provide tools to calculate Local Parameter Importance
==================================================================================

.. module:: orion.analysis.lpi
   :platform: Unix
   :synopsis: Provide tools to calculate Local Parameter Importance
"""
import numpy
import pandas as pd
from sklearn.ensemble import AdaBoostRegressor, BaggingRegressor,\
    ExtraTreesRegressor, GradientBoostingRegressor, RandomForestRegressor

from orion.core.worker.transformer import build_required_space


_regressors_ = {
    'AdaBoostRegressor': AdaBoostRegressor,
    'BaggingRegressor': BaggingRegressor,
    'ExtraTreesRegressor': ExtraTreesRegressor,
    'GradientBoostingRegressor': GradientBoostingRegressor,
    'RandomForestRegressor': RandomForestRegressor,
}


def train_regressor(regressor_name, data, **kwargs):
    """Train regressor model

    Parameters
    ----------
    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)
    trials: DataFrame or dict
        A dataframe of trials containing, at least, the columns 'objective' and 'id'. Or a dict
        equivalent.

    **kwargs
        Arguments for the regressor model.

    """
    if regressor_name not in _regressors_:
        raise ValueError(
            f'{regressor_name} is not a supported regressor. '
            f'Did you mean any of theses: list(_regressors_.keys())')

    regressor = _regressors_[regressor_name](**kwargs)
    return regressor.fit(data[:, :-1], data[:, -1])


def to_numpy(trials, space):
    """Convert trials in DataFrame to Numpy array of (params + objective)"""
    return trials[list(space.keys()) + ['objective']].to_numpy()


def flatten(trials_array, flattened_space):
    """Flatten dimensions"""
    flattened_points = numpy.array(
        [flattened_space.transform(point[:-1]) for point in trials_array])

    return numpy.concatenate((flattened_points, trials_array[:, -1:]), axis=1)


def make_grid(point, space, model, n):
    """Build a grid based on point.

    The shape of the grid will be
        (number of hyperparameters,
         number of points ``n``,
         number of hyperparameters + 1)

    Last column is the objective predicted by the model for a given point.

    Parameters
    ----------
    point: numpy.ndarray
        A tuple representation of the best trials, (hyperparameters + objective)
    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)
    trials: DataFrame or dict
        A dataframe of trials containing, at least, the columns 'objective' and 'id'. Or a dict
        equivalent.

    **kwargs
        Arguments for the regressor model.

    """
    grid = numpy.zeros((len(space), n, len(space) + 1))
    for i, dim in enumerate(space.values()):
        grid[i, :, :] = point
        grid[i, :, i] = numpy.linspace(*dim.interval(), num=n)
        grid[i, :, -1] = model.predict(grid[i, :, :-1])
    return grid


def compute_variances(grid):
    """Compute variance for each hyperparameters"""
    return grid[:, :, -1].var(axis=1)


def _lpi(point, space, model, n):
    """Local parameter importance for each hyperparameters"""
    grid = make_grid(point, space, model, n)
    variances = compute_variances(grid)
    ratios = variances / variances.sum()
    return pd.DataFrame(data=ratios, index=space.keys(), columns=['LPI'])


def _linear_lpi(point, space, model, n):
    # TODO
    return


modes = dict(
    best=_lpi,
    linear=_linear_lpi)


def lpi(trials, space, mode='best', model='RandomForestRegressor', n=20, **kwargs):
    """
    Calculates the Local Parameter Importance for a collection of :class:`Trial`.

    For more information on the metric, see original paper at
    https://ml.informatik.uni-freiburg.de/papers/18-LION12-CAVE.pdf.

    Biedenkapp, André, et al. "Cave: Configuration assessment, visualization and evaluation."
    International Conference on Learning and Intelligent Optimization. Springer, Cham, 2018.

    Parameters
    ----------
    trials: DataFrame or dict
        A dataframe of trials containing, at least, the columns 'objective' and 'id'. Or a dict
        equivalent.

    space: Space object
        A space object from an experiment.

    mode: str
        Mode to compute the LPI.
        - ``best``: Take the best trial found as the anchor for the LPI
        - ``linear``: Recompute LPI for all values on a grid

    model: str
        Name of the regression model to use. Can be one of
        - AdaBoostRegressor
        - BaggingRegressor
        - ExtraTreesRegressor
        - GradientBoostingRegressor
        - RandomForestRegressor (Default)

    n: int
        Number of points to compute the variances. Default is 20.

    **kwargs
        Arguments for the regressor model.

    Returns
    -------
    DataFrame
        LPI value for each parameter. If ``mode`` is `linear`, then a list of
        param values and LPI metrics are returned in a DataFrame format.
    """
    flattened_space = build_required_space(
        space, type_requirement='numerical', shape_requirement='flattened')
    if trials.empty or trials.shape[0] == 0:
        return pd.DataFrame(
            data=[0] * len(flattened_space),
            index=flattened_space.keys(),
            columns=['LPI'])

    data = to_numpy(trials, space)
    data = flatten(data, flattened_space)
    model = train_regressor(model, data, **kwargs)
    best_point = data[numpy.argmin(data[:, -1])]
    results = modes[mode](best_point, flattened_space, model, n)
    return results
